import unittest
from unittest import mock
from betfairlightweight.resources.bettingresources import PriceSize

from flumine.order.order import OrderStatus, OrderTypes, BetdaqOrder
from flumine import config
from flumine.markets.market import Market
from flumine.markets.markets import Markets
from flumine.order.order import (
    BaseOrder,
    BetfairOrder,
)
from flumine.order import process
from flumine.strategy.strategy import Strategies
from flumine.utils import create_cheap_hash


class ProcessCurrentOrdersTest(unittest.TestCase):
    @mock.patch("flumine.order.process.process_current_order")
    def test_process_current_orders_with_default_sep(self, mock_process_current_order):
        mock_log_control = mock.Mock()
        mock_add_market = mock.Mock()
        market_book = mock.Mock()
        markets = Markets()
        market = Market(
            flumine=mock.Mock(), market_id="market_id", market_book=market_book
        )
        markets.add_market("market_id", market)
        strategies = Strategies()
        cheap_hash = create_cheap_hash("strategy_name", 13)
        trade = mock.Mock(market_id="market_id")
        trade.strategy.name_hash = cheap_hash
        current_order = mock.Mock(
            customer_order_ref=f"{cheap_hash}I123", market_id="market_id", bet_id=None
        )
        betfair_order = BetfairOrder(trade=trade, side="BACK", order_type=mock.Mock())
        betfair_order.id = "123"
        betfair_order.complete = True
        market.blotter["123"] = betfair_order
        event = mock.Mock(event=[mock.Mock(orders=[current_order])])

        process.process_current_orders(
            markets=markets,
            strategies=strategies,
            event=event,
            log_control=mock_log_control,
            add_market=mock_add_market,
        )
        mock_process_current_order.assert_called_with(
            betfair_order, current_order, mock_log_control
        )
        self.assertEqual(market.blotter._live_orders, [])

    def test_process_current_order(self):
        mock_order = mock.Mock(status=OrderStatus.EXECUTABLE)
        mock_order.current_order.status = "EXECUTION_COMPLETE"
        mock_current_order = mock.Mock()
        mock_log_control = mock.Mock()
        process.process_current_order(mock_order, mock_current_order, mock_log_control)
        mock_order.update_current_order.assert_called_with(mock_current_order)
        mock_order.execution_complete.assert_called()

    @mock.patch("flumine.order.process.OrderEvent")
    def test_process_current_order_async(self, mock_order_event):
        mock_order = mock.Mock(status=OrderStatus.EXECUTABLE, async_=True, bet_id=None)
        mock_order.current_order.status = "EXECUTION_COMPLETE"
        mock_current_order = mock.Mock(bet_id=1234)
        mock_log_control = mock.Mock()
        process.process_current_order(mock_order, mock_current_order, mock_log_control)
        mock_order.update_current_order.assert_called_with(mock_current_order)
        mock_order.execution_complete.assert_called()
        self.assertEqual(mock_order.bet_id, 1234)
        mock_order.responses.placed.assert_called_with()
        mock_order_event.assert_called_with(mock_order, venue=mock_order.VENUE)
        mock_log_control.assert_called_with(mock_order_event())

    def test_create_order_from_current(self):
        mock_add_market = mock.Mock()
        market_book = mock.Mock()
        mock_client = mock.Mock()
        mock_flumine = mock.Mock()
        markets = Markets()
        market = Market(
            flumine=mock.Mock(), market_id="market_id", market_book=market_book
        )
        markets.add_market("market_id", market)
        cheap_hash = create_cheap_hash("strategy_name", 13)
        strategy = mock.Mock(name_hash=cheap_hash)
        strategies = Strategies()
        strategies(strategy=strategy, clients=mock.Mock(), flumine=mock_flumine)
        current_order = mock.Mock(
            customer_order_ref=f"{cheap_hash}I123",
            market_id="market_id",
            bet_id=None,
            selection_id="selection_id",
            handicap="handicap",
            order_type="LIMIT",
            price_size=PriceSize(price=10.0, size=2.0),
            persistence_type="LAPSE",
        )

        new_order = process.create_order_from_current(
            markets=markets,
            strategies=strategies,
            current_order=current_order,
            add_market=mock_add_market,
            client=mock_client,
        )
        self.assertEqual(market.blotter["123"], new_order)
        self.assertEqual(new_order.market_id, "market_id")
        self.assertEqual(new_order.selection_id, "selection_id")
        self.assertEqual(new_order.handicap, "handicap")
        self.assertEqual(new_order.order_type.ORDER_TYPE, OrderTypes.LIMIT)
        self.assertEqual(new_order.order_type.size, 2.0)
        self.assertEqual(new_order.order_type.price, 10.0)
        self.assertEqual(new_order.client, mock_client)

    @mock.patch("flumine.order.process.process_betdaq_current_order")
    def test_process_betdaq_current_orders(self, mock_process_betdaq_current_order):
        mock_log_control = mock.Mock()
        mock_add_market = mock.Mock()
        market_book = mock.Mock()
        markets = Markets()
        market = Market(
            flumine=mock.Mock(), market_id="market_id", market_book=market_book
        )
        markets.add_market("market_id", market)
        strategies = Strategies()
        cheap_hash = create_cheap_hash("strategy_name", 13)
        trade = mock.Mock(market_id="market_id")
        trade.strategy.name_hash = cheap_hash
        current_order = mock.Mock(
            customer_order_ref=f"{cheap_hash}I123", market_id="market_id", bet_id=None
        )
        current_order = {"order_id": 456, "customer_reference": 123}
        betdaq_order = BetdaqOrder(trade=trade, side="BACK", order_type=mock.Mock())
        betdaq_order.id = "123"
        betdaq_order.complete = True
        market.blotter["123"] = betdaq_order
        event = mock.Mock(event=[current_order])

        process.process_betdaq_current_orders(
            markets=markets,
            strategies=strategies,
            event=event,
            log_control=mock_log_control,
            add_market=mock_add_market,
        )
        mock_process_betdaq_current_order.assert_called_with(
            betdaq_order, current_order
        )
        self.assertEqual(market.blotter._live_orders, [])

    def test_process_betdaq_current_order_pending(self):
        mock_order = mock.Mock(
            bet_id=123,
            status=OrderStatus.PENDING,
            current_order={"status": "Unmatched"},
        )
        mock_current_order = mock.Mock()
        process.process_betdaq_current_order(mock_order, mock_current_order)
        mock_order.update_current_order.assert_called_with(mock_current_order)
        mock_order.executable.assert_called()

    def test_process_betdaq_current_order_pending_matched(self):
        mock_order = mock.Mock(
            bet_id=123, status=OrderStatus.PENDING, current_order={"status": "Matched"}
        )
        mock_current_order = mock.Mock()
        process.process_betdaq_current_order(mock_order, mock_current_order)
        mock_order.update_current_order.assert_called_with(mock_current_order)
        mock_order.execution_complete.assert_called()

    def test_process_betdaq_current_order_matched(self):
        mock_order = mock.Mock(
            status=OrderStatus.EXECUTABLE, current_order={"status": "Matched"}
        )
        mock_current_order = mock.Mock()
        process.process_betdaq_current_order(mock_order, mock_current_order)
        mock_order.update_current_order.assert_called_with(mock_current_order)
        mock_order.execution_complete.assert_called()

    def test_process_betdaq_current_order_update(self):
        mock_order = mock.Mock(
            status=OrderStatus.UPDATING,
            current_order={"sequence_number": 1, "status": "Unmatched"},
        )
        mock_current_order = {"sequence_number": 2, "price": 999}
        process.process_betdaq_current_order(mock_order, mock_current_order)
        mock_order.update_current_order.assert_called_with(mock_current_order)
        mock_order.executable.assert_called()

    def test_process_betdaq_current_order_update_matched(self):
        mock_order = mock.Mock(
            status=OrderStatus.UPDATING,
            current_order={"sequence_number": 1, "status": "Matched"},
        )
        mock_current_order = {"sequence_number": 2, "price": 999, "status": "Matched"}
        process.process_betdaq_current_order(mock_order, mock_current_order)
        mock_order.update_current_order.assert_called_with(mock_current_order)
        mock_order.executable.assert_not_called()
        mock_order.execution_complete.assert_called()

    def test_process_betdaq_current_order_update_race(self):
        mock_order = mock.Mock(
            status=OrderStatus.UPDATING, current_order={"sequence_number": 1}
        )
        mock_current_order = {"sequence_number": 1}
        process.process_betdaq_current_order(mock_order, mock_current_order)
        mock_order.executable.assert_not_called()



import datetime
from enum import Enum

from flumine.clients import VenueType


class EventType(Enum):
    CONFIG = "Config"
    TERMINATOR = "Terminator"
    # betfair objects
    MARKET_CATALOGUE = "MarketCatalogue"
    MARKET_BOOK = "MarketBook"
    SPORTS_DATA = "SportsData"
    RAW_DATA = "Raw streaming data"
    CURRENT_ORDERS = "CurrentOrders"
    CLEARED_MARKETS = "ClearedMarkets"
    CLEARED_ORDERS = "ClearedOrders"
    CLEARED_ORDERS_META = "ClearedOrders metadata"
    BALANCE = "Balance"
    # flumine objects
    STRATEGY = "Strategy"
    MARKET = "Market"
    TRADE = "Trade"
    ORDER = "Order"
    ORDER_PACKAGE = "Order package"
    CLOSE_MARKET = "Closed market"
    CUSTOM_EVENT = "Custom event"


class QueueType(Enum):
    HANDLER = "Handler queue"
    LOGGING = "Logging queue"


class BaseEvent:
    EVENT_TYPE = None
    QUEUE_TYPE = None

    __slots__ = ["_time_created", "event", "venue", "callback"]

    def __init__(self, event, venue: VenueType = VenueType.BETFAIR):
        self._time_created = datetime.datetime.now(datetime.timezone.utc)
        self.event = event
        self.venue = venue

    @property
    def elapsed_seconds(self):
        return (
            datetime.datetime.now(datetime.timezone.utc) - self._time_created
        ).total_seconds()

    def __str__(self):
        return "<{0} [{1}]>".format(self.EVENT_TYPE.name, self.QUEUE_TYPE.name)


# HANDLER


class MarketCatalogueEvent(BaseEvent):
    EVENT_TYPE = EventType.MARKET_CATALOGUE
    QUEUE_TYPE = QueueType.HANDLER


class MarketBookEvent(BaseEvent):
    EVENT_TYPE = EventType.MARKET_BOOK
    QUEUE_TYPE = QueueType.HANDLER


class SportsDataEvent(BaseEvent):
    EVENT_TYPE = EventType.SPORTS_DATA
    QUEUE_TYPE = QueueType.HANDLER


class RawDataEvent(BaseEvent):
    EVENT_TYPE = EventType.RAW_DATA
    QUEUE_TYPE = QueueType.HANDLER


class CurrentOrdersEvent(BaseEvent):
    EVENT_TYPE = EventType.CURRENT_ORDERS
    QUEUE_TYPE = QueueType.HANDLER

    def __init__(
        self,
        event,
        venue: VenueType = VenueType.BETFAIR,
        callback=None,
    ):
        super(CurrentOrdersEvent, self).__init__(event, venue)
        self.callback = callback


class ClearedMarketsEvent(BaseEvent):
    EVENT_TYPE = EventType.CLEARED_MARKETS
    QUEUE_TYPE = QueueType.HANDLER


class ClearedOrdersEvent(BaseEvent):
    EVENT_TYPE = EventType.CLEARED_ORDERS
    QUEUE_TYPE = QueueType.HANDLER


class CloseMarketEvent(BaseEvent):
    EVENT_TYPE = EventType.CLOSE_MARKET
    QUEUE_TYPE = QueueType.HANDLER


class CustomEvent(BaseEvent):
    EVENT_TYPE = EventType.CUSTOM_EVENT
    QUEUE_TYPE = QueueType.HANDLER

    def __init__(self, event, callback, *args, **kwargs):
        super(CustomEvent, self).__init__(event)
        self.callback = callback


# LOGGING


class ConfigEvent(BaseEvent):
    EVENT_TYPE = EventType.CONFIG
    QUEUE_TYPE = QueueType.LOGGING


class ClearedOrdersMetaEvent(BaseEvent):
    EVENT_TYPE = EventType.CLEARED_ORDERS_META
    QUEUE_TYPE = QueueType.LOGGING


class BalanceEvent(BaseEvent):
    EVENT_TYPE = EventType.BALANCE
    QUEUE_TYPE = QueueType.LOGGING


class StrategyEvent(BaseEvent):
    EVENT_TYPE = EventType.STRATEGY
    QUEUE_TYPE = QueueType.LOGGING


class MarketEvent(BaseEvent):
    EVENT_TYPE = EventType.MARKET
    QUEUE_TYPE = QueueType.LOGGING


class TradeEvent(BaseEvent):
    EVENT_TYPE = EventType.TRADE
    QUEUE_TYPE = QueueType.LOGGING


class OrderEvent(BaseEvent):
    EVENT_TYPE = EventType.ORDER
    QUEUE_TYPE = QueueType.LOGGING


# both


class TerminationEvent(BaseEvent):
    EVENT_TYPE = EventType.TERMINATOR
    QUEUE_TYPE = QueueType.HANDLER



import time
import queue
import logging
import threading
from typing import Type
from betfairlightweight import resources

from .strategy.strategy import Strategies, BaseStrategy
from .streams.streams import Streams
from .events import events
from .worker import BackgroundWorker
from .clients import Clients, BaseClient, VenueType
from .markets.markets import Markets
from .markets.market import Market
from .markets.middleware import Middleware, SimulatedMiddleware
from .execution.betfairexecution import BetfairExecution
from .execution.simulatedexecution import SimulatedExecution
from .execution.betdaqexecution import BetdaqExecution
from .order.process import process_current_orders, process_betdaq_current_orders
from .controls.clientcontrols import BaseControl, MaxTransactionCount
from .controls.tradingcontrols import (
    OrderValidation,
    StrategyExposure,
    MarketValidation,
)
from .controls.loggingcontrols import LoggingControl
from .exceptions import FlumineException, ClientError
from . import config, utils

logger = logging.getLogger(__name__)


class BaseFlumine:
    SIMULATED = False

    def __init__(self, client: BaseClient = None):
        """
        Base framework class

        :param client: flumine client instance
        """
        self._running = False
        # streams (market/order)
        self.streams = Streams(self)

        self.clients = Clients()

        # FIFO queue
        self.handler_queue = queue.Queue()

        # markets
        self.markets = Markets()

        # middleware
        self._market_middleware = []

        # strategies
        self.strategies = Strategies()

        # order execution class
        self.simulated_execution = SimulatedExecution(self)
        self.betfair_execution = BetfairExecution(self)
        self.betdaq_execution = BetdaqExecution(self, 1)

        # add client
        if client:
            self.add_client(client)

        # logging controls (e.g. database logger)
        self._logging_controls = []

        # trading controls
        self.trading_controls = []
        # add default controls (processed in order)
        self.add_trading_control(OrderValidation)
        self.add_trading_control(MarketValidation)
        self.add_trading_control(StrategyExposure)

        # workers
        self._workers = []

    def run(self) -> None:
        raise NotImplementedError

    def add_client(self, client: BaseClient) -> None:
        self.clients.add_client(client)
        self.streams.add_client(client)
        # add execution
        client.add_execution(self)
        # add simulation middleware if required
        if self.clients.simulated and not any(
            isinstance(val, SimulatedMiddleware) for val in self._market_middleware
        ):
            self.add_market_middleware(SimulatedMiddleware())
        # register default client controls (processed in order)
        self.add_client_control(client, MaxTransactionCount)

    def add_strategy(self, strategy: BaseStrategy) -> None:
        logger.info("Adding strategy %s", strategy)
        self.streams(strategy)  # create required streams
        self.strategies(strategy, self.clients, self)  # store in strategies
        self.log_control(events.StrategyEvent(strategy))

    def add_worker(self, worker: BackgroundWorker) -> None:
        logger.info("Adding worker %s", worker.name)
        self._workers.append(worker)

    def add_client_control(
        self, client: BaseClient, client_control: Type[BaseControl], **kwargs
    ) -> None:
        logger.info("Adding client control %s", client_control.NAME)
        client.trading_controls.append(client_control(self, client, **kwargs))

    def add_trading_control(self, trading_control: Type[BaseControl], **kwargs) -> None:
        logger.info("Adding trading control %s", trading_control.NAME)
        self.trading_controls.append(trading_control(self, **kwargs))

    def add_market_middleware(self, middleware: Middleware) -> None:
        logger.info("Adding market middleware %s", middleware)
        self._market_middleware.append(middleware)

    def add_logging_control(self, logging_control: LoggingControl) -> None:
        logger.info("Adding logging control %s", logging_control.NAME)
        self._logging_controls.append(logging_control)

    def log_control(self, event: events.BaseEvent) -> None:
        for logging_control in self._logging_controls:
            logging_control.logging_queue.put(event)

    def _add_default_workers(self) -> None:
        return

    def _process_market_books(self, event: events.MarketBookEvent) -> None:
        for market_book in event.event:
            market_id = market_book.market_id

            # check latency (only if marketBook is from a stream update)
            if market_book.streaming_snap is False:
                latency = time.time() - (market_book.publish_time_epoch / 1e3)
                if latency > 2:
                    logger.warning(
                        "High latency between current time and MarketBook publish time",
                        extra={
                            "market_id": market_id,
                            "latency": latency,
                            "pt": market_book.publish_time,
                        },
                    )

            market = self.markets.markets.get(market_id)
            market_is_new = market is None
            if market_is_new:
                market = self._add_market(market_id, market_book)
            elif market.closed:
                self.markets.add_market(market_id, market)

            if market_book.status == "CLOSED":
                self.handler_queue.put(events.CloseMarketEvent(market_book))
                continue

            # process market
            market(market_book)

            # process middleware
            for middleware in self._market_middleware:
                utils.call_middleware_error_handling(middleware, market)

            for strategy in self.strategies:
                if market_book.streaming_unique_id in strategy.stream_ids:
                    if market_is_new:
                        utils.call_strategy_error_handling(
                            strategy.process_new_market, market, market_book
                        )
                    if utils.call_strategy_error_handling(
                        strategy.check_market_book, market, market_book
                    ):
                        utils.call_strategy_error_handling(
                            strategy.process_market_book, market, market_book
                        )

    def _process_sports_data(self, event: events.SportsDataEvent) -> None:
        for sports_data in event.event:
            if hasattr(sports_data, "event_id"):
                # get eventId
                event_id = sports_data.event_id
                # get markets
                markets = self.markets.events[event_id]
            else:
                market_id = sports_data.market_id
                market = self.markets.markets.get(market_id)
                if market is None:
                    logger.error(
                        "Market not present for sports data",
                        extra={"market_id": market_id},
                    )
                    continue
                markets = [market]
            # loop markets
            for market in markets:
                for strategy in self.strategies:
                    if sports_data.streaming_unique_id in strategy.stream_ids:
                        if utils.call_strategy_error_handling(
                            strategy.check_sports_data, market, sports_data
                        ):
                            utils.call_strategy_error_handling(
                                strategy.process_sports_data, market, sports_data
                            )

    def process_order_package(self, order_package) -> None:
        """Execute through client."""
        order_package.client.execution.handler(order_package)

    def _add_market(self, market_id: str, market_book: resources.MarketBook) -> Market:
        logger.info("Adding: %s to markets", market_id)
        market = Market(self, market_id, market_book)
        self.markets.add_market(market_id, market)
        for middleware in self._market_middleware:
            middleware.add_market(market)
        return market

    def _remove_market(self, market: Market, clear: bool = True) -> None:
        logger.info("Removing market %s", market.market_id, extra=self.info)
        for middleware in self._market_middleware:
            middleware.remove_market(market)
        for strategy in self.strategies:
            strategy.remove_market(market.market_id)
        if clear:
            self.markets.remove_market(market.market_id)

    def _process_raw_data(self, event: events.RawDataEvent) -> None:
        stream_id, clk, publish_time, data = event.event
        for datum in data:
            if "id" in datum:
                market_id = datum["id"]
                market = self.markets.markets.get(market_id)
                if market is None:
                    market = self._add_market(market_id, None)
                elif market.closed:
                    self.markets.add_market(market_id, market)

                if "marketDefinition" in datum:
                    market.update_market_catalogue = True
                    if datum["marketDefinition"]["status"] == "CLOSED":
                        datum["_stream_id"] = stream_id
                        self.handler_queue.put(events.CloseMarketEvent(datum))

            for strategy in self.strategies:
                if stream_id in strategy.stream_ids:
                    utils.call_process_raw_data(strategy, clk, publish_time, datum)

    def _process_market_catalogues(self, event: events.MarketCatalogueEvent) -> None:
        for market_catalogue in event.event:
            market = self.markets.markets.get(market_catalogue.market_id)
            if market:
                if market.market_catalogue is None:
                    market.market_catalogue = market_catalogue
                    self.log_control(events.MarketEvent(market))
                    logger.info(
                        "Created marketCatalogue for %s",
                        market.market_id,
                        extra=market.info,
                    )
                else:
                    market.market_catalogue = market_catalogue
                    logger.info(
                        "Updated marketCatalogue for %s",
                        market.market_id,
                        extra=market.info,
                    )
                market.update_market_catalogue = False

                for strategy in self.strategies:
                    if (
                        market.market_book
                        and market.market_book.streaming_unique_id
                        in strategy.stream_ids
                    ) or strategy.market_cached(market.market_id):
                        utils.call_strategy_error_handling(
                            strategy.process_market_catalogue, market, market_catalogue
                        )

    def _process_current_orders(self, event: events.CurrentOrdersEvent) -> None:
        # update state
        if event.event:
            if event.callback:
                event.callback(
                    self.markets,
                    self.strategies,
                    event,
                    self.log_control,
                    self._add_market,
                )
            else:
                if event.venue == VenueType.BETFAIR:
                    process_current_orders(
                        self.markets,
                        self.strategies,
                        event,
                        self.log_control,
                        self._add_market,
                    )
                elif event.venue == VenueType.BETDAQ:
                    process_betdaq_current_orders(
                        self.markets,
                        self.strategies,
                        event,
                        self.log_control,
                        self._add_market,
                    )
        for market in self.markets:
            if market.closed is False and market.blotter.active:
                for strategy in self.strategies:
                    strategy_orders = market.blotter.strategy_orders(strategy)
                    if strategy_orders:
                        utils.call_process_orders_error_handling(
                            strategy, market, strategy_orders
                        )

    def _process_custom_event(self, event: events.CustomEvent) -> None:
        try:
            event.callback(self, event)
        except FlumineException as e:
            logger.error(
                "FlumineException error %s in _process_custom_event %s",
                e,
                event.callback,
                exc_info=True,
            )
        except Exception as e:
            logger.exception(
                "Unknown error %s in _process_custom_event %s",
                e,
                event.callback,
                exc_info=True,
            )
            if config.raise_errors:
                raise

    def _process_close_market(self, event: events.CloseMarketEvent) -> None:
        market_book = event.event
        if isinstance(market_book, dict):
            recorder = True
            market_id = market_book["id"]
            stream_id = market_book["_stream_id"]
        else:
            recorder = False
            market_id = market_book.market_id
            stream_id = market_book.streaming_unique_id
        market = self.markets.markets.get(market_id)
        if market is None:
            logger.warning(
                "Market %s not present when closing",
                market_id,
                extra={"market_id": market_id, **self.info},
            )
            return
        # process market
        if market.closed is False:
            market.close_market()
        if recorder is False:
            market(market_book)
            market.blotter.process_closed_market(market, event.event)

        for strategy in self.strategies:
            if stream_id in strategy.stream_ids or strategy.market_filter == {}:
                strategy.process_closed_market(market, event.event)

        if recorder is False and self.clients.simulated:
            # simulate ClearedOrdersEvent
            cleared_orders = resources.ClearedOrders(
                moreAvailable=False, clearedOrders=[]
            )
            cleared_orders.market_id = market_id
            self._process_cleared_orders(events.ClearedOrdersEvent(cleared_orders))
            for client in self.clients:
                # simulate ClearedMarketsEvent
                cleared_markets = resources.ClearedOrders(
                    moreAvailable=False,
                    clearedOrders=[market.cleared(client)],
                )
                self._process_cleared_markets(
                    events.ClearedMarketsEvent(cleared_markets)
                )
        self.log_control(event)
        logger.info("Market closed", extra={"market_id": market_id, **self.info})

        # check for markets that have been closed for x seconds and remove
        if not self.clients.simulated:
            # due to monkey patching this will clear simulated markets
            closed_markets = [
                m
                for m in self.markets
                if m.closed
                and m.elapsed_seconds_closed
                and m.elapsed_seconds_closed > 3600
            ]
            for market in closed_markets:
                self._remove_market(market)
        else:
            self._remove_market(market, clear=False)

    def _process_cleared_orders(self, event):
        if event.venue == VenueType.BETFAIR:
            market_id = event.event.market_id
            market = self.markets.markets.get(market_id)
            if market is None:
                logger.warning(
                    "Market %s not present when clearing" % market_id,
                    extra={"market_id": market_id, **self.info},
                )
                return

            meta_orders = market.blotter.process_cleared_orders(event.event)
        elif event.venue == VenueType.BETDAQ:
            market_id = "betdaq"
            meta_orders = []
            for cleared_order in event.event:
                ref = cleared_order["customer_reference"]
                # check every market for the order
                for market in self.markets:
                    try:
                        order = market.blotter[str(ref)]
                    except KeyError:
                        continue
                    if order:
                        order.cleared_order = cleared_order
                        meta_orders.append(order)
                        break
                else:
                    continue
        else:
            logger.warning("Unknown venue in '_process_cleared_orders'")
            return
        if meta_orders:
            self.log_control(events.ClearedOrdersMetaEvent(meta_orders))
        logger.info(
            "Market cleared",
            extra={
                "market_id": market_id,
                "order_count": len(meta_orders),
                **self.info,
            },
        )

    def _process_cleared_markets(self, event: events.ClearedMarketsEvent):
        # todo update blotter?
        for cleared_market in event.event.orders:
            logger.info(
                "Market level cleared",
                extra={
                    "market_id": cleared_market.market_id,
                    "profit": cleared_market.profit,
                    "bet_count": cleared_market.bet_count,
                },
            )
        self.log_control(event)

    def _process_end_flumine(self) -> None:
        self.strategies.finish(self)

    @property
    def info(self) -> dict:
        return {
            "clients": self.clients.info,
            "markets": {
                "market_count": len(self.markets),
                "open_market_count": len(self.markets.open_market_ids),
            },
            "streams": [s for s in self.streams],
            "logging_controls": self._logging_controls,
            "threads": threading.enumerate(),
        }

    def __enter__(self):
        logger.info("Starting flumine", extra=self.info)
        if len(self.clients) == 0:
            raise ClientError("No clients provided")
        # simulated
        if self.SIMULATED:
            config.simulated = True
        else:
            config.simulated = False
        # login
        self.clients.login()
        self.clients.update_account_details()
        # add default and start all workers
        self._add_default_workers()
        for w in self._workers:
            w.start()
        # start logging controls
        for c in self._logging_controls:
            c.start()
        # process config (logging)
        self.log_control(events.ConfigEvent(config))
        # start strategies
        self.strategies.start(self)
        # start streams
        self.streams.start()

        self._running = True

    def __exit__(self, *args):
        # shutdown framework
        self._process_end_flumine()
        # shutdown workers
        for w in self._workers:
            w.shutdown()
        # shutdown streams
        self.streams.stop()
        # shutdown thread pools
        self.simulated_execution.shutdown()
        self.betfair_execution.shutdown()
        self.betdaq_execution.shutdown()
        # shutdown logging controls
        self.log_control(events.TerminationEvent(self))
        for c in self._logging_controls:
            if c.is_alive():
                c.join()
        # logout
        self.clients.logout()
        self._running = False
        logger.info("Exiting flumine", extra=self.info)



import logging
from typing import Optional

from ..markets.markets import Markets
from ..order.order import BaseOrder, OrderStatus
from ..order.trade import Trade
from ..strategy.strategy import Strategies
from ..events.events import OrderEvent
from ..utils import STRATEGY_NAME_HASH_LENGTH

logger = logging.getLogger(__name__)

"""
Various functions to update current order status
and update status.

Loop through each current order:
    order = Lookup order in market using marketId and orderId
    if order is None (not present locally):
        create local order using data and make executable #todo!!!
    if order betId != current_order betId:
        Get order using current_order betId due to replace request (new betId)
    if order:
        process()
            update current status
            if betId is None (async placement):
                Update betId and log through logging_control

orderStatus: PENDING, EXECUTION_COMPLETE, EXECUTABLE, EXPIRED
"""


def process_current_orders(
    markets: Markets, strategies: Strategies, event, log_control, add_market
) -> None:
    for current_orders in event.event:
        client = current_orders.client
        for current_order in current_orders.orders:
            if current_order.customer_order_ref is None:
                continue
            order_id = current_order.customer_order_ref[STRATEGY_NAME_HASH_LENGTH + 1 :]
            order = markets.get_order(
                market_id=current_order.market_id,
                order_id=order_id,
            )
            if order is None:
                logger.warning(
                    "Order %s not present in blotter" % current_order.bet_id,
                    extra={
                        "bet_id": current_order.bet_id,
                        "market_id": current_order.market_id,
                        "customer_strategy_ref": current_order.customer_strategy_ref,
                        "customer_order_ref": current_order.customer_order_ref,
                        "client_username": client.username,
                    },
                )
                order = create_order_from_current(
                    markets, strategies, current_order, add_market, client
                )
                if order is None:
                    continue

            if (
                order.bet_id and order.bet_id != current_order.bet_id
            ):  # replaceOrder handling (hacky)
                order = markets.get_order_from_bet_id(
                    market_id=current_order.market_id,
                    bet_id=current_order.bet_id,
                )
                if order is None:
                    continue
            # process order status
            process_current_order(order, current_order, log_control)
            # complete order if required
            if order.complete:
                market = markets.markets[order.market_id]
                if order in market.blotter.live_orders:
                    market.blotter.complete_order(order)


def process_current_order(order: BaseOrder, current_order, log_control) -> None:
    # update
    order.update_current_order(current_order)
    # pickup async orders
    if order.async_ and order.bet_id is None and current_order.bet_id:
        order.responses.placed()
        order.bet_id = current_order.bet_id
        log_control(OrderEvent(order, venue=order.VENUE))
    # update status
    if order.bet_id and order.status == OrderStatus.PENDING:
        if order.current_order.status == "EXECUTABLE":
            order.executable()
        elif order.current_order.status in ["EXECUTION_COMPLETE", "EXPIRED"]:
            order.execution_complete()
    elif order.status == OrderStatus.EXECUTABLE:
        if order.current_order.status in ["EXECUTION_COMPLETE", "EXPIRED"]:
            order.execution_complete()


def create_order_from_current(
    markets: Markets, strategies: Strategies, current_order, add_market, client
) -> Optional[BaseOrder]:
    strategy_name_hash = current_order.customer_order_ref[:STRATEGY_NAME_HASH_LENGTH]
    order_id = current_order.customer_order_ref[STRATEGY_NAME_HASH_LENGTH + 1 :]
    # get strategy
    strategy = strategies.hashes.get(strategy_name_hash)
    if strategy is None:
        logger.warning(
            "Strategy not available to create order %s",
            order_id,
            extra={
                "bet_id": current_order.bet_id,
                "market_id": current_order.market_id,
                "customer_strategy_ref": current_order.customer_strategy_ref,
                "customer_order_ref": current_order.customer_order_ref,
                "strategy_name_hash": strategy_name_hash,
            },
        )
        return
    # get market
    market = markets.markets.get(current_order.market_id)
    if market is None:
        # create market
        market = add_market(current_order.market_id, market_book=None)
    # add trade/order
    trade = Trade(
        market.market_id, current_order.selection_id, current_order.handicap, strategy
    )
    order = trade.create_order_from_current(client, current_order, order_id)
    market.blotter[order.id] = order
    runner_context = strategy.get_runner_context(*order.lookup)
    runner_context.place(trade.id)
    order.placing()
    logger.info(
        "process: New order trade created",
        extra={
            "bet_id": current_order.bet_id,
            "market_id": current_order.market_id,
            "customer_strategy_ref": current_order.customer_strategy_ref,
            "customer_order_ref": current_order.customer_order_ref,
            "strategy_name": str(strategy),
            "client_username": client.username,
        },
    )
    return order


def process_betdaq_current_orders(
    markets: Markets, strategies: Strategies, event, log_control, add_market
) -> None:
    for current_order in event.event:
        ref = current_order["customer_reference"]
        # check every market for the order
        for market in markets:
            try:
                order = market.blotter[str(ref)]
            except KeyError:
                continue
            if order:
                break
        else:
            continue
        # process order status
        process_betdaq_current_order(order, current_order)
        # complete order if required
        if order.complete:
            for market in markets:
                if order in market.blotter.live_orders:
                    market.blotter.complete_order(order)
                    break


def process_betdaq_current_order(order: BaseOrder, current_order) -> None:
    old_sequence_number = order.current_order.get("sequence_number")
    # update
    order.update_current_order(current_order)
    # todo pick up NoReceipt orders
    # update status
    if order.status == OrderStatus.PENDING and order.bet_id:
        if order.current_order["status"] in ["Unmatched", "Suspended"]:
            order.executable()
        else:
            order.execution_complete()
    elif (
        order.status == OrderStatus.UPDATING
        and old_sequence_number != current_order.get("sequence_number")
    ):
        order.order_type.price = current_order["price"]
        if order.current_order["status"] in ["Unmatched", "Suspended"]:
            order.executable()
        else:
            order.execution_complete()
    elif order.status == OrderStatus.EXECUTABLE:
        if order.current_order["status"] in ["Matched", "Cancelled", "Settled", "Void"]:
            order.execution_complete()



import time
import requests
from typing import Optional

from .baseexecution import BaseExecution
from .. import config
from ..clients.clients import VenueType
from ..order.orderpackage import BaseOrderPackage, OrderPackageType


class SimulatedExecution(BaseExecution):
    VENUE = VenueType.SIMULATED

    def handler(self, order_package: BaseOrderPackage) -> None:
        """Only uses _thread_pool if paper_trade"""
        if order_package.package_type == OrderPackageType.PLACE:
            func = self.execute_place
        elif order_package.package_type == OrderPackageType.CANCEL:
            func = self.execute_cancel
        elif order_package.package_type == OrderPackageType.UPDATE:
            func = self.execute_update
        elif order_package.package_type == OrderPackageType.REPLACE:
            func = self.execute_replace
        else:
            raise NotImplementedError()

        if order_package.client.paper_trade:
            self._thread_pool.submit(func, order_package, None)
        else:
            func(order_package, http_session=None)

    def execute_place(
        self, order_package, http_session: Optional[requests.Session]
    ) -> None:
        if order_package.client.paper_trade:
            time.sleep(order_package.bet_delay + config.place_latency)
        market = self.flumine.markets.markets[order_package.market_id]
        for order, instruction in zip(order_package, order_package.place_instructions):
            with order.trade:
                self._bet_id += 1
                simulated_response = order.simulated.place(
                    order_package, market.market_book, instruction, self._bet_id
                )
                self._order_logger(
                    order, simulated_response, order_package.package_type
                )
                if simulated_response.status == "SUCCESS":
                    order.executable()
                elif simulated_response.status == "FAILURE":
                    order.execution_complete()

        # update transaction counts
        order_package.client.add_transaction(len(order_package))

    def execute_cancel(
        self, order_package, http_session: Optional[requests.Session]
    ) -> None:
        if order_package.client.paper_trade:
            time.sleep(config.cancel_latency)
        market = self.flumine.markets.markets[order_package.market_id]
        failed_transaction_count = 0
        for order in order_package:
            with order.trade:
                simulated_response = order.simulated.cancel(market.market_book)
                self._order_logger(
                    order, simulated_response, order_package.package_type
                )
                if simulated_response.status == "SUCCESS":
                    if order.size_remaining == 0:
                        order.execution_complete()
                    else:
                        order.executable()
                elif simulated_response.status == "FAILURE":
                    order.executable()
                    failed_transaction_count += 1

        # update transaction counts
        if failed_transaction_count:
            order_package.client.add_transaction(failed_transaction_count, failed=True)

    def execute_update(
        self, order_package, http_session: Optional[requests.Session]
    ) -> None:
        if order_package.client.paper_trade:
            time.sleep(config.update_latency)
        market = self.flumine.markets.markets[order_package.market_id]
        failed_transaction_count = 0
        for order, instruction in zip(order_package, order_package.update_instructions):
            with order.trade:
                simulated_response = order.simulated.update(
                    market.market_book, instruction
                )
                self._order_logger(
                    order, simulated_response, order_package.package_type
                )
                if simulated_response.status == "SUCCESS":
                    order.executable()
                elif simulated_response.status == "FAILURE":
                    order.executable()
                    failed_transaction_count += 1

        # update transaction counts
        if failed_transaction_count:
            order_package.client.add_transaction(failed_transaction_count, failed=True)

    def execute_replace(
        self, order_package, http_session: Optional[requests.Session]
    ) -> None:
        if (
            order_package.client.paper_trade
        ):  # todo should the cancel happen without a delay?
            time.sleep(order_package.bet_delay + config.replace_latency)
        market = self.flumine.markets.markets[order_package.market_id]
        failed_transaction_count = 0
        for order, instruction in zip(
            order_package, order_package.replace_instructions
        ):
            with order.trade:
                # cancel current order
                err = False
                cancel_instruction_report = order.simulated.cancel(market.market_book)
                if cancel_instruction_report.status == "SUCCESS":
                    order.execution_complete()
                elif cancel_instruction_report.status == "FAILURE":
                    order.executable()
                    failed_transaction_count += 1
                    err = True
                self._order_logger(
                    order,
                    cancel_instruction_report,
                    OrderPackageType.CANCEL,
                )
                if err:
                    continue

                # place new order
                self._bet_id += 1
                replacement_order = order.trade.create_order_replacement(
                    order,
                    instruction.get("newPrice"),
                    cancel_instruction_report.size_cancelled,
                    order_package.date_time_created,
                )
                place_instruction_report = replacement_order.simulated.place(
                    order_package, market.market_book, instruction, self._bet_id
                )
                if place_instruction_report.status == "SUCCESS":
                    self._order_logger(
                        replacement_order,
                        place_instruction_report,
                        order_package.package_type,
                    )
                    # add to blotter
                    market.place_order(
                        replacement_order, execute=False, client=order.client
                    )
                    replacement_order.executable()
                elif place_instruction_report.status == "FAILURE":
                    order.executable()

        # update transaction counts
        order_package.client.add_transaction(len(order_package))
        if failed_transaction_count:
            order_package.client.add_transaction(failed_transaction_count, failed=True)



import unittest
from unittest import mock

from flumine.strategy import strategy
from flumine.strategy.runnercontext import RunnerContext


class RunnerContextMock(RunnerContext):
    """
    A mock class for RunnerContext which maintains most of its properties
    except placed_elapsed_seconds and reset_elapsed_seconds, which are replaced
    with attributes.
    """

    # Properties cannot be replaced with attributes on instances because the
    # assignment invokes a setter, but they can be replaced on a class level
    placed_elapsed_seconds = None
    reset_elapsed_seconds = None

    def __init__(
        self,
        completed_trades: list = [],
        live_trades: list = [],
        placed_elapsed_seconds: float = None,
        reset_elapsed_seconds: float = None,
    ):
        super().__init__(12345678)  # Use dummy selection id
        all_trades = completed_trades + live_trades
        if len(set(all_trades)) != len(all_trades):
            raise ValueError("All trade ids must be unique.")
        for trade_id in all_trades:
            self.place(trade_id)
        for trade_id in completed_trades:
            self.reset(trade_id)
        self.placed_elapsed_seconds = placed_elapsed_seconds
        self.reset_elapsed_seconds = reset_elapsed_seconds


class StrategiesTest(unittest.TestCase):
    def setUp(self) -> None:
        self.strategies = strategy.Strategies()

    def test_init(self):
        self.assertEqual(self.strategies._strategies, [])

    def test_call(self):
        mock_strategy = mock.Mock()
        mock_clients = mock.Mock()
        mock_flumine = mock.Mock()
        self.strategies(mock_strategy, mock_clients, mock_flumine)
        self.assertEqual(self.strategies._strategies, [mock_strategy])
        mock_strategy.add.assert_called_with(mock_flumine)
        self.assertEqual(mock_strategy.clients, mock_clients)

    def test_start(self):
        mock_strategy = mock.Mock()
        mock_flumine = mock.Mock()
        self.strategies._strategies.append(mock_strategy)
        self.strategies.start(mock_flumine)
        mock_strategy.start.assert_called_with(mock_flumine)

    def test_finish(self):
        mock_flumine = mock.Mock()
        mock_strategy = mock.Mock()
        self.strategies._strategies.append(mock_strategy)
        self.strategies.finish(mock_flumine)
        mock_strategy.finish.assert_called_with(mock_flumine)

    def test_iter(self):
        for i in self.strategies:
            assert i

    def test_len(self):
        self.assertEqual(len(self.strategies), 0)


class BaseStrategyTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_market_filter = mock.Mock()
        self.mock_market_data_filter = mock.Mock()
        self.streaming_timeout = 2
        self.conflate_ms = 100
        self.strategy = strategy.BaseStrategy(
            market_filter=self.mock_market_filter,
            market_data_filter=self.mock_market_data_filter,
            sports_data_filter=["cricketSubscription"],
            streaming_timeout=self.streaming_timeout,
            conflate_ms=self.conflate_ms,
            stream_class=strategy.MarketStream,
            name="test",
            context={"trigger": 0.123},
            max_selection_exposure=1,
            max_order_exposure=2,
            max_trade_count=5,
            max_live_trade_count=3,
            multi_order_trades=False,
        )

    def test_init(self):
        self.assertEqual(self.strategy.market_filter, self.mock_market_filter)
        self.assertEqual(self.strategy.market_data_filter, self.mock_market_data_filter)
        self.assertEqual(self.strategy.sports_data_filter, ["cricketSubscription"])
        self.assertEqual(self.strategy.streaming_timeout, self.streaming_timeout)
        self.assertEqual(self.strategy.conflate_ms, self.conflate_ms)
        self.assertEqual(self.strategy.stream_class, strategy.MarketStream)
        self.assertEqual(self.strategy._name, "test")
        self.assertEqual(self.strategy.context, {"trigger": 0.123})
        self.assertEqual(self.strategy.max_selection_exposure, 1)
        self.assertEqual(self.strategy.max_order_exposure, 2)
        self.assertIsNone(self.strategy.clients)
        self.assertEqual(self.strategy.max_trade_count, 5)
        self.assertEqual(self.strategy.max_live_trade_count, 3)
        self.assertEqual(self.strategy.streams, [])
        self.assertEqual(self.strategy.historic_stream_ids, set())
        self.assertEqual(self.strategy.name_hash, "a94a8fe5ccb19")
        self.assertFalse(self.strategy.multi_order_trades)
        self.assertEqual(strategy.STRATEGY_NAME_HASH_LENGTH, 13)
        self.assertEqual(
            strategy.DEFAULT_MARKET_DATA_FILTER,
            {
                "fields": [
                    "EX_ALL_OFFERS",
                    "EX_TRADED",
                    "EX_TRADED_VOL",
                    "EX_LTP",
                    "EX_MARKET_DEF",
                    "SP_TRADED",
                    "SP_PROJECTED",
                ]
            },
        )

    def test_add(self):
        mock_flumine = mock.Mock()
        self.strategy.add(mock_flumine)

    def test_start(self):
        mock_flumine = mock.Mock()
        self.strategy.start(mock_flumine)

    def test_process_new_market(self):
        self.strategy.process_new_market(None, None)

    def test_check_market_book(self):
        self.assertFalse(self.strategy.check_market_book(None, None))

    def test_process_market_book(self):
        self.strategy.process_market_book(None, None)

    def test_process_market_catalogue(self):
        self.strategy.process_market_catalogue(None, None)

    def test_check_sports_data(self):
        self.assertFalse(self.strategy.check_sports_data(None, None))

    def test_process_sports_data(self):
        self.strategy.process_sports_data(None, None)

    def test_process_raw_data(self):
        self.strategy.process_raw_data(None, None, None)

    def test_process_orders(self):
        self.strategy.process_orders(None, None)

    def test_process_closed_market(self):
        self.strategy.process_closed_market(None, None)

    def test_finish(self):
        self.strategy.finish(mock.Mock())

    def test_remove_market(self):
        self.strategy._invested = {
            ("1.23", 456, 7): 1,
            ("1.23", 891, 7): 2,
            ("1.24", 112, 7): 3,
        }
        self.strategy.remove_market("1.23")
        self.assertEqual(self.strategy._invested, {("1.24", 112, 7): 3})

    def test_validate_order(self):
        mock_order = mock.Mock()
        mock_order.trade.id = "99"  # Unused trade id, nonexistent in runner_context
        # No orders/trades placed yet
        runner_context = RunnerContextMock()
        self.strategy.log_validation = True
        self.assertTrue(self.strategy.validate_order(runner_context, mock_order))
        # trade count (5 is limit)
        runner_context = RunnerContextMock(["1", "2", "3", "4", "5"])
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        # live trade count (3 is limit)
        runner_context = RunnerContextMock([], ["3", "4", "5"])
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        # place elapsed
        runner_context = RunnerContextMock(
            live_trades=["1"], placed_elapsed_seconds=0.49
        )
        mock_order.trade.place_reset_seconds = 0.5
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        # reset elapsed
        runner_context = RunnerContextMock(
            completed_trades=["1"], placed_elapsed_seconds=1, reset_elapsed_seconds=0.49
        )
        mock_order.trade.reset_seconds = 0.5
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))

    def test_validate_order_reuse_completed_trade(self):
        """Reusing a completed Trade object to place a new order."""
        mock_order = mock.Mock()
        mock_order.trade.id = "1"  # Reuse completed trade
        # trade count - at the limit (5)
        runner_context = RunnerContextMock(["1", "2", "3", "4", "5"])
        self.assertTrue(self.strategy.validate_order(runner_context, mock_order))
        # live trade count- at the limit (3)
        runner_context = RunnerContextMock(["1"], ["3", "4", "5"])
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        # trade count - over the limit (>5) (force=True was used)
        runner_context = RunnerContextMock(["1", "2", "3", "4", "5", "6"])
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        # live trade count- over the limit (>3) (force=True was used)
        runner_context = RunnerContextMock(["1"], ["3", "4", "5", "6"])
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))

    def test_validate_order_reuse_live_trade(self):
        """Reusing a live Trade object to place a new order."""
        mock_order = mock.Mock()
        mock_order.trade.id = "5"  # Reuse live trade
        # trade count - at the limit (5)
        runner_context = RunnerContextMock(["1", "2", "3"], ["4", "5"])
        self.assertTrue(self.strategy.validate_order(runner_context, mock_order))
        # live trade count - at the limit (3)
        runner_context = RunnerContextMock([], ["3", "4", "5"])
        self.assertTrue(self.strategy.validate_order(runner_context, mock_order))
        # trade count - over the limit (>5) (force=True was used)
        runner_context = RunnerContextMock(["1", "2", "3", "4"], ["5", "6"])
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        # live trade count- over the limit (>3) (force=True was used)
        runner_context = RunnerContextMock([], ["3", "4", "5", "6"])
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))

    def test_validate_order_multi(self):
        mock_order = mock.Mock()
        mock_order.trade.id = "6"  # In live trades
        runner_context = RunnerContextMock([], ["1", "2", "3", "4", "5", "6"])
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        self.strategy.multi_order_trades = True
        self.assertTrue(self.strategy.validate_order(runner_context, mock_order))

    def test_executable_orders(self):
        mock_context = mock.Mock(executable_orders=True)
        self.strategy._invested = {("2", 456, 1): mock_context}
        self.assertFalse(self.strategy.has_executable_orders("1", 123, 1.0))
        self.assertFalse(self.strategy.has_executable_orders("2", 123, 1.0))
        self.assertTrue(self.strategy.has_executable_orders("2", 456, 1.0))

    def test_get_runner_context(self):
        mock_context = mock.Mock(invested=True)
        self.strategy._invested = {("2", 456, 0): mock_context}
        self.assertEqual(self.strategy.get_runner_context("2", 456, 0), mock_context)
        self.strategy.get_runner_context("2", 789, 0)
        self.assertEqual(len(self.strategy._invested), 2)

    def test_market_cached(self):
        # This is not a good test as it relies too much on implementation details.
        # There is an integration test as well to cover a more realistic use-case.
        mock_stream = mock.Mock()
        mock_stream._listener.stream._caches = {"1.234": None}
        self.strategy.streams = [mock_stream]
        self.assertTrue(self.strategy.market_cached("1.234"))
        self.assertFalse(self.strategy.market_cached("1.789"))

    def test_market_cached_stream_not_registered(self):
        """
        Tests that strategy.market_cached() does not throw an error when called
        on a stream which had not been registered yet
        (betfairlightweight.BaseListener.register_stream not called).
        """
        mock_stream = mock.Mock()
        mock_stream._listener.stream = None
        self.strategy.streams = [mock_stream]
        self.assertFalse(self.strategy.market_cached("1.234"))
        self.assertFalse(self.strategy.market_cached("1.789"))

    def test_stream_ids(self):
        mock_stream = mock.Mock(stream_id=321)
        self.strategy.streams = [mock_stream]
        self.assertEqual(self.strategy.stream_ids, [321])
        self.strategy.historic_stream_ids.add(123)
        self.assertEqual(self.strategy.stream_ids, {123})

    def test_info(self):
        self.assertEqual(
            self.strategy.info,
            {
                "conflate_ms": self.conflate_ms,
                "market_data_filter": self.mock_market_data_filter,
                "market_filter": self.mock_market_filter,
                "strategy_name": "test",
                "name_hash": "a94a8fe5ccb19",
                "stream_ids": [],
                "streaming_timeout": self.streaming_timeout,
                "context": {"trigger": 0.123},
                "max_live_trade_count": 3,
                "max_order_exposure": 2,
                "max_selection_exposure": 1,
                "max_trade_count": 5,
            },
        )

    def test_name(self):
        self.assertEqual(self.strategy.name, "test")

    def test_str(self):
        self.assertEqual(str(self.strategy), "test")



import time
import threading
import logging
from typing import Callable, Optional
from betdaq import BetdaqError
from betfairlightweight import BetfairError, filters, exceptions

from . import config
from .events import events
from .utils import chunks
from .clients import VenueType

logger = logging.getLogger(__name__)


class BackgroundWorker(threading.Thread):
    def __init__(
        self,
        flumine,
        function: Callable,
        interval: Optional[int],
        func_args: tuple = None,
        func_kwargs: dict = None,
        start_delay: int = 0,
        context: dict = None,
        name: str = None,
        **kwargs,
    ):
        name = name or function.__name__
        threading.Thread.__init__(self, daemon=True, name=name, **kwargs)
        self.flumine = flumine
        self.function = function
        self.interval = interval
        self.func_args = func_args if func_args is not None else []
        self.func_kwargs = func_kwargs if func_kwargs is not None else {}
        self.start_delay = start_delay
        self.context = context or {}
        self._running = False

    def run(self) -> None:
        logger.info(
            "BackgroundWorker %s starting",
            self.name,
            extra={
                "worker_name": self.name,
                "function": self.function,
                "context": self.context,
                "start_delay": self.start_delay,
                "interval": self.interval,
                "func_args": self.func_args,
                "func_kwargs": self.func_kwargs,
            },
        )
        time.sleep(self.start_delay)
        self._running = True
        while self._running:
            logger.debug(
                "BackgroundWorker %s executing",
                self.name,
                extra={
                    "worker_name": self.name,
                    "function": self.function,
                    "context": self.context,
                },
            )
            try:
                self.function(
                    self.context, self.flumine, *self.func_args, **self.func_kwargs
                )
            except Exception as e:
                logger.error(
                    "Error in BackgroundWorker %s: %s",
                    self.name,
                    e,
                    extra={
                        "worker_name": self.name,
                        "function": self.function,
                        "context": self.context,
                    },
                    exc_info=True,
                )
            if self.interval is None:
                break
            time.sleep(self.interval)

    def shutdown(self, timeout: int = 4) -> None:
        logger.info(
            "BackgroundWorker %s shutting down",
            self.name,
            extra={
                "worker_name": self.name,
                "function": self.function,
            },
        )
        self._running = False
        self.join(timeout)


def keep_alive(context: dict, flumine) -> None:
    """Attempt keep alive if required or
    login if keep alive failed
    """
    for client in flumine.clients:
        if client.VENUE == VenueType.BETFAIR:
            if client.betting_client.session_token:
                resp = client.keep_alive()
                if resp is True or resp.status == "SUCCESS":
                    continue
        elif client.VENUE == VenueType.BETCONNECT:
            resp = client.keep_alive()
            if resp:
                continue
        elif client.VENUE == VenueType.BETDAQ:
            continue
        # keep-alive failed lets try a login
        client.login()


def poll_market_catalogue(context: dict, flumine) -> None:
    # get betfair client
    client = flumine.clients.get_default(VenueType.BETFAIR)
    markets = [
        m.market_id
        for m in list(flumine.markets.markets.values())
        if m.update_market_catalogue and not m.closed
    ]
    for market_ids in chunks(markets, 25):
        try:
            market_catalogues = client.betting_client.betting.list_market_catalogue(
                filter=filters.market_filter(market_ids=market_ids),
                max_results=25,
                market_projection=[
                    "COMPETITION",
                    "EVENT",
                    "EVENT_TYPE",
                    "RUNNER_DESCRIPTION",
                    "RUNNER_METADATA",
                    "MARKET_START_TIME",
                    "MARKET_DESCRIPTION",
                ],
            )
        except exceptions.StatusCodeError as e:
            # log as warning to prevent duplicate logs on betfair meltdown
            logger.warning(
                "poll_market_catalogue StatusCodeError",
                extra={"trading_function": "list_market_catalogue", "response": e},
                exc_info=True,
            )
            continue
        except BetfairError as e:
            logger.error(
                "poll_market_catalogue error",
                exc_info=True,
                extra={"trading_function": "list_market_catalogue", "response": e},
            )
            continue

        if market_catalogues:
            flumine.handler_queue.put(events.MarketCatalogueEvent(market_catalogues))


def poll_account_balance(context: dict, flumine) -> None:
    for client in flumine.clients:
        client.update_account_details()
        logger.info("Client update account details", extra=client.info)
        if client.account_funds:
            flumine.log_control(events.BalanceEvent(client, venue=client.VENUE))


def poll_market_closure(context: dict, flumine) -> None:
    markets = [
        market for market in list(flumine.markets.markets.values()) if market.closed
    ]
    for client in flumine.clients:
        if (
            client.VENUE != VenueType.BETFAIR
            or client.paper_trade
            or client.market_recording_mode
        ):
            continue
        for market in markets:
            if client.username not in market.orders_cleared:
                if _get_cleared_orders(
                    flumine, client.betting_client, market.market_id
                ):
                    market.orders_cleared.append(client.username)
            if client.username not in market.market_cleared:
                if _get_cleared_market(
                    flumine, client.betting_client, market.market_id
                ):
                    market.market_cleared.append(client.username)


def _get_cleared_orders(flumine, betting_client, market_id: str) -> bool:
    from_record = 0
    while True:
        try:
            cleared_orders = betting_client.betting.list_cleared_orders(
                bet_status="SETTLED",
                from_record=from_record,
                market_ids=[market_id],
                customer_strategy_refs=(
                    [config.customer_strategy_ref]
                    if config.customer_strategy_ref
                    else None
                ),
            )
        except exceptions.StatusCodeError as e:
            # log as warning to prevent duplicate logs on betfair meltdown
            logger.warning(
                "_get_cleared_orders StatusCodeError",
                extra={"trading_function": "list_cleared_orders", "response": e},
                exc_info=True,
            )
            return False
        except BetfairError as e:
            logger.warning(
                "_get_cleared_orders error",
                extra={"trading_function": "list_cleared_orders", "response": e},
                exc_info=True,
            )
            return False

        logger.info(
            "%s: %s cleared orders found, more available: %s",
            market_id,
            len(cleared_orders.orders),
            cleared_orders.more_available,
        )

        from_record += len(cleared_orders.orders)
        if from_record == 0:
            market = flumine.markets.markets[market_id]
            flumine_size_matched = sum([order.size_matched for order in market.blotter])
            if flumine_size_matched > 0:
                return False

        cleared_orders.market_id = market_id
        flumine.handler_queue.put(events.ClearedOrdersEvent(cleared_orders))
        flumine.log_control(events.ClearedOrdersEvent(cleared_orders))

        if not cleared_orders.more_available:
            break
    return True


def _get_cleared_market(flumine, betting_client, market_id: str) -> bool:
    try:
        cleared_markets = betting_client.betting.list_cleared_orders(
            bet_status="SETTLED",
            market_ids=[market_id],
            customer_strategy_refs=(
                [config.customer_strategy_ref] if config.customer_strategy_ref else None
            ),
            group_by="MARKET",
        )
    except exceptions.StatusCodeError as e:
        # log as warning to prevent duplicate logs on betfair meltdown
        logger.warning(
            "_get_cleared_market StatusCodeError",
            extra={"trading_function": "list_cleared_orders", "response": e},
            exc_info=True,
        )
        return False
    except BetfairError as e:
        logger.error(
            "_get_cleared_market error",
            extra={"trading_function": "list_cleared_orders", "response": e},
            exc_info=True,
        )
        return False

    if cleared_markets.orders:
        flumine.handler_queue.put(events.ClearedMarketsEvent(cleared_markets))
        return True
    else:
        return False


def betdaq_settled_orders(context: dict, flumine) -> None:
    for client in flumine.clients:
        if client.VENUE == VenueType.BETDAQ:
            sequence_number = context.get(client, 0)
            # get initial sequence number
            if sequence_number == 0:
                current_orders = client.betting_client.betting.get_orders(
                    sequence_number
                )
                sequence_number = current_orders["maximum_sequence_number"]
                context[client] = sequence_number

            while True:
                try:
                    current_orders = client.betting_client.betting.get_orders_diff(
                        sequence_number
                    )
                except (BetdaqError, Exception):
                    time.sleep(1)
                    logger.error("betdaq_settled_orders run error", exc_info=True)
                    continue

                if not current_orders:
                    break
                cleared_orders = [
                    o
                    for o in current_orders
                    if o["status"] in ["Settled", "Cancelled", "Void"]
                ]
                if cleared_orders:
                    flumine.handler_queue.put(
                        events.ClearedOrdersEvent(
                            cleared_orders, venue=VenueType.BETDAQ
                        )
                    )
                    flumine.log_control(
                        events.ClearedOrdersEvent(
                            cleared_orders, venue=VenueType.BETDAQ
                        )
                    )
                # update SequenceNumber
                new_sequence_number = max(
                    sequence_number,
                    *(
                        o.get("sequence_number", sequence_number)
                        for o in current_orders
                    ),
                )
                if new_sequence_number == sequence_number:
                    break

                sequence_number = new_sequence_number
                context[client] = new_sequence_number



from typing import Optional
from betfairlightweight.metadata import transaction_limit as betfair_transaction_limit

from ..utils import create_short_uuid
from .clients import VenueType

DEFAULT_CAPITAL_BASE = 0
DEFAULT_COMMISSION_BASE = 0.05


class BaseClient:
    """
    Abstraction of betting client.
    """

    VENUE = None

    def __init__(
        self,
        betting_client=None,
        transaction_limit: Optional[int] = betfair_transaction_limit,
        capital_base: int = DEFAULT_CAPITAL_BASE,
        commission_base: float = DEFAULT_COMMISSION_BASE,
        interactive_login: bool = False,
        username: str = None,
        order_stream: bool = True,
        order_stream_conflate_ms: int = None,
        best_price_execution: bool = True,
        min_bet_validation: bool = True,
        paper_trade: bool = False,
        market_recording_mode: bool = False,
        simulated_full_match: bool = False,
        execution_cls=None,
        order_stream_cls=None,
    ):
        if hasattr(betting_client, "lightweight"):
            assert (
                betting_client.lightweight is False
            ), "flumine requires resources, please set lightweight to False"
        self._username = username or create_short_uuid()
        self.betting_client = betting_client
        self.transaction_limit = transaction_limit
        self.capital_base = capital_base
        self.commission_base = commission_base  # not implemented
        self.interactive_login = interactive_login
        self.order_stream = order_stream
        self.order_stream_conflate_ms = order_stream_conflate_ms
        self.best_price_execution = best_price_execution  # simulation only
        self.min_bet_validation = min_bet_validation  # used in OrderValidation control
        self.paper_trade = paper_trade  # simulated order placement using live data
        self.market_recording_mode = market_recording_mode  # no order stream / workers
        self.simulated_full_match = (
            simulated_full_match  # simulated 100% match on successful place
        )

        self.account_details = None
        self.account_funds = None
        self.commission_paid = 0

        self._execution_cls = execution_cls  # custom execution
        self.order_stream_cls = order_stream_cls  # custom order stream
        self.execution = None  # set during flumine init
        self.trading_controls = []

    def login(self) -> None:
        raise NotImplementedError

    def keep_alive(self) -> None:
        raise NotImplementedError

    def logout(self) -> None:
        raise NotImplementedError

    def update_account_details(self) -> None:
        raise NotImplementedError

    def add_execution(self, flumine) -> None:
        if self._execution_cls:
            self.execution = self._execution_cls(flumine)
        else:
            if self.VENUE == VenueType.SIMULATED or self.paper_trade:
                self.execution = flumine.simulated_execution
            elif self.VENUE == VenueType.BETFAIR:
                self.execution = flumine.betfair_execution
            elif self.VENUE == VenueType.BETDAQ:
                self.execution = flumine.betdaq_execution

    def add_transaction(self, count: int, failed: bool = False) -> None:
        for control in self.trading_controls:
            if hasattr(control, "add_transaction"):
                control.add_transaction(count, failed)

    @property
    def username(self) -> str:
        if self.betting_client:
            return self.betting_client.username
        else:
            return self._username

    @property
    def current_transaction_count_total(self) -> Optional[int]:
        # current hours total transaction count
        for control in self.trading_controls:
            if control.NAME == "MAX_TRANSACTION_COUNT":
                return control.current_transaction_count_total

    @property
    def transaction_count_total(self) -> Optional[int]:
        # total transaction count
        for control in self.trading_controls:
            if control.NAME == "MAX_TRANSACTION_COUNT":
                return control.transaction_count_total

    @property
    def min_bet_size(self) -> Optional[float]:
        raise NotImplementedError

    @property
    def min_bet_payout(self) -> Optional[float]:
        raise NotImplementedError

    @property
    def min_bsp_liability(self) -> Optional[float]:
        raise NotImplementedError

    @property
    def info(self) -> dict:
        return {
            "username": self.username,
            "venue": self.VENUE.value if self.VENUE else None,
            "betting_client": self.betting_client,
            "current_transaction_count_total": self.current_transaction_count_total,
            "transaction_count_total": self.transaction_count_total,
            "trading_controls": self.trading_controls,
            "order_stream": self.order_stream,
            "order_stream_conflate_ms": self.order_stream_conflate_ms,
            "best_price_execution": self.best_price_execution,
            "paper_trade": self.paper_trade,
        }



import datetime
import logging
import collections
import unittest
from unittest import mock

from flumine.order.trade import Trade, OrderError, TradeStatus


class TradeTest(unittest.TestCase):
    def setUp(self) -> None:
        logging.disable(logging.CRITICAL)
        mock_client = mock.Mock(paper_trade=False)
        self.mock_strategy = mock.Mock(client=mock_client)
        self.notes = collections.OrderedDict({"trigger": 123})
        self.trade = Trade(
            "1.234",
            567,
            1.0,
            self.mock_strategy,
            self.notes,
            12,
            34,
        )

    def test_init(self):
        self.assertEqual(self.trade.market_id, "1.234")
        self.assertEqual(self.trade.selection_id, 567)
        self.assertEqual(self.trade.handicap, 1.0)
        self.assertEqual(self.trade.strategy, self.mock_strategy)
        self.assertEqual(self.trade.notes, self.notes)
        self.assertEqual(self.trade.status_log, [])
        self.assertEqual(self.trade.orders, [])
        self.assertFalse(self.trade.pending_orders)
        self.assertIsNotNone(self.trade.date_time_created)
        self.assertIsNone(self.trade.date_time_complete)
        self.assertIsNone(self.trade.market_notes)
        self.assertEqual(self.trade.place_reset_seconds, 12)
        self.assertEqual(self.trade.reset_seconds, 34)

    @mock.patch("flumine.order.trade.Trade.complete_trade")
    @mock.patch(
        "flumine.order.trade.Trade.complete",
        new_callable=mock.PropertyMock,
        return_value=True,
    )
    def test__update_status(self, mock_complete, mock_complete_trade):
        self.trade._update_status(TradeStatus.COMPLETE)
        self.assertEqual(self.trade.status_log, [TradeStatus.COMPLETE])
        self.assertEqual(self.trade.status, TradeStatus.COMPLETE)
        mock_complete_trade.assert_called()

    @mock.patch("flumine.order.trade.Trade._update_status")
    def test_complete_trade(self, mock__update_status):
        self.trade.complete_trade()
        mock__update_status.assert_called_with(TradeStatus.COMPLETE)
        runner_context = self.mock_strategy.get_runner_context(
            self.trade.market_id, self.trade.selection_id, self.trade.handicap
        )
        runner_context.reset.assert_called_with(self.trade.id)
        self.assertIsNotNone(self.trade.date_time_complete)

    def test_complete(self):
        self.assertTrue(self.trade.complete)
        mock_order = mock.Mock(complete=False)
        self.trade.orders.append(mock_order)
        self.assertFalse(self.trade.complete)

    def test_trade_complete_pending(self):
        self.trade.pending_orders = True
        self.assertFalse(self.trade.complete)
        self.trade.pending_orders = False
        self.assertTrue(self.trade.complete)

    def test_trade_complete_replace_order(self):
        self.assertTrue(self.trade.complete)
        mock_order = mock.Mock(complete=True)
        self.trade.status = TradeStatus.COMPLETE
        self.trade.orders.append(mock_order)
        self.assertFalse(self.trade.complete)

    def test_create_order(self):
        mock_order_type = mock.Mock(VENUE="SYM")
        mock_order = mock.Mock(VENUE="SYM")
        self.trade.create_order(
            "BACK", mock_order_type, order=mock_order, sep="-", context={1: 2}
        )
        mock_order.assert_called_with(
            trade=self.trade,
            side="BACK",
            order_type=mock_order_type,
            handicap=self.trade.handicap,
            sep="-",
            context={1: 2},
            notes=None,
        )
        self.assertEqual(self.trade.orders, [mock_order()])

    def test_create_order_error(self):
        mock_order_type = mock.Mock(VENUE="SYM")
        mock_order = mock.Mock(VENUE="MYS")
        with self.assertRaises(OrderError):
            self.trade.create_order("BACK", mock_order_type, order=mock_order)

    def test_create_order_replacement(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        mock_order = mock.Mock(sep="-")
        replacement_order = self.trade.create_order_replacement(
            mock_order, 12, 2.00, now
        )
        self.assertEqual(self.trade.orders, [replacement_order])
        self.assertEqual(replacement_order.trade, self.trade)
        self.assertEqual(replacement_order.side, mock_order.side)
        self.assertEqual(replacement_order.order_type.price, 12)
        self.assertEqual(replacement_order.order_type.size, 2.00)
        self.assertEqual(
            replacement_order.order_type.persistence_type,
            mock_order.order_type.persistence_type,
        )
        self.assertEqual(replacement_order.handicap, mock_order.handicap)
        self.assertEqual(replacement_order.sep, mock_order.sep)
        self.assertEqual(replacement_order.context, mock_order.context)
        self.assertEqual(replacement_order.notes, mock_order.notes)
        self.assertEqual(replacement_order.client, mock_order.client)
        self.assertEqual(replacement_order.date_time_created, now)
        self.assertEqual(replacement_order.responses.date_time_created, now)

    def test_create_order_from_current_limit(self):
        mock_client = mock.Mock()
        mock_current_order = mock.Mock(
            order_type="LIMIT",
            placed_date=datetime.datetime.fromtimestamp(12, tz=datetime.timezone.utc),
            matched_date=None,
            cancelled_date=datetime.datetime.fromtimestamp(
                34, tz=datetime.timezone.utc
            ),
        )
        order = self.trade.create_order_from_current(
            mock_client, mock_current_order, "12345"
        )
        self.assertEqual(order.bet_id, mock_current_order.bet_id)
        self.assertEqual(order.id, "12345")
        self.assertEqual(order.client, mock_client)
        self.assertEqual(self.trade.orders, [order])
        self.assertEqual(
            order.date_time_created,
            datetime.datetime.fromtimestamp(12, tz=datetime.timezone.utc),
        )
        self.assertEqual(
            order.date_time_execution_complete,
            datetime.datetime.fromtimestamp(34, tz=datetime.timezone.utc),
        )

    def test_create_order_from_current_limit_on_close(self):
        mock_client = mock.Mock()
        mock_current_order = mock.Mock(order_type="LIMIT_ON_CLOSE")
        order = self.trade.create_order_from_current(
            mock_client, mock_current_order, "12345"
        )
        self.assertEqual(order.bet_id, mock_current_order.bet_id)
        self.assertEqual(order.id, "12345")
        self.assertEqual(order.client, mock_client)
        self.assertEqual(self.trade.orders, [order])

    def test_create_order_from_current_market_on_close(self):
        mock_client = mock.Mock()
        mock_current_order = mock.Mock(order_type="MARKET_ON_CLOSE")
        order = self.trade.create_order_from_current(
            mock_client, mock_current_order, "12345"
        )
        self.assertEqual(order.bet_id, mock_current_order.bet_id)
        self.assertEqual(order.id, "12345")
        self.assertEqual(order.client, mock_client)
        self.assertEqual(self.trade.orders, [order])

    def test_create_order_from_current_unknown(self):
        mock_client = mock.Mock()
        mock_current_order = mock.Mock(order_type="BE")
        with self.assertRaises(NotImplementedError):
            self.trade.create_order_from_current(
                mock_client, mock_current_order, "12345"
            )

    def test_create_betdaq_order(self):
        mock_order_type = mock.Mock(VENUE="BETDAQ")
        mock_order = mock.Mock(VENUE="BETDAQ")
        self.trade.create_order(
            "BACK", mock_order_type, order=mock_order, sep="-", context={1: 2}
        )
        mock_order.assert_called_with(
            trade=self.trade,
            side="BACK",
            order_type=mock_order_type,
            handicap=self.trade.handicap,
            sep="-",
            context={1: 2},
            notes=None,
        )
        self.assertEqual(self.trade.orders, [mock_order()])

    def test_elapsed_seconds(self):
        mock_order_one = mock.Mock(elapsed_seconds=123)
        mock_order_two = mock.Mock(elapsed_seconds=12)
        self.trade.orders = [mock_order_one, mock_order_two]
        self.assertGreaterEqual(self.trade.elapsed_seconds, 123)

    def test_notes_str(self):
        self.trade.notes = collections.OrderedDict({"1": 1, 2: "2", 3: 3, 4: "four"})
        self.assertEqual(self.trade.notes_str, "1,2,3,four")
        self.trade.notes = collections.OrderedDict()
        self.assertEqual(self.trade.notes_str, "")

    def test_info(self):
        self.trade.status_log = [TradeStatus.PENDING, TradeStatus.COMPLETE]
        self.assertEqual(
            self.trade.info,
            {
                "id": str(self.trade.id),
                "orders": [],
                "pending_orders": False,
                "place_reset_seconds": 12,
                "reset_seconds": 34,
                "strategy": str(self.mock_strategy),
                "notes": "123",
                "market_notes": None,
                "status": "Live",
                "status_log": "Pending, Complete",
            },
        )

    @mock.patch("flumine.order.trade.Trade._update_status")
    def test_enter_exit(self, mock__update_status):
        with self.trade:
            mock__update_status.assert_called_with(TradeStatus.PENDING)
        mock__update_status.assert_called_with(TradeStatus.LIVE)

    @mock.patch("flumine.order.trade.Trade._update_status")
    def test_enter_error(self, mock__update_status):
        with self.assertRaises(ValueError):
            with self.trade:
                raise ValueError()
        mock__update_status.assert_called_with(TradeStatus.PENDING)



import os
import logging
from collections import defaultdict
from betfairlightweight.resources.bettingresources import RunnerBook

from ..order.order import OrderStatus, OrderTypes
from ..utils import wap, call_strategy_error_handling
from ..streams.historicalstream import (
    HistoricListener,
    FlumineHistoricalGeneratorStream,
)
from .. import config

logger = logging.getLogger(__name__)

WIN_MINIMUM_ADJUSTMENT_FACTOR = 2.5
PLACE_MINIMUM_ADJUSTMENT_FACTOR = 0  # todo implement correctly (https://en-betfair.custhelp.com/app/answers/detail/a_id/406)
LIVE_STATUS = [
    OrderStatus.EXECUTABLE,
    OrderStatus.CANCELLING,
    OrderStatus.UPDATING,
    OrderStatus.REPLACING,
]


class Middleware:
    def __call__(self, market) -> None:
        pass

    def add_market(self, market) -> None:
        pass

    def remove_market(self, market) -> None:
        pass


class SimulatedMiddleware(Middleware):
    """
    Calculates matched amounts per runner
    to be used in simulated matching.
    # todo currency fluctuations fucks everything
    """

    def __init__(self):
        # {marketId: {(selectionId, handicap): RunnerAnalytics}}
        self.markets = defaultdict(dict)
        self._runner_removals = []

    def __call__(self, market) -> None:
        market_analytics = self.markets[market.market_id]
        runner_removals = []  # [(selectionId, handicap, adjustmentFactor)..]

        for runner in market.market_book.runners:
            if runner.status == "ACTIVE":
                self._process_runner(market_analytics, runner)
            elif runner.status == "REMOVED":
                _removal = (
                    runner.selection_id,
                    runner.handicap,
                    runner.adjustment_factor,
                )
                if _removal not in self._runner_removals:
                    logger.warning(
                        "Runner %s (%s) removed from market %s",
                        runner.selection_id,
                        runner.adjustment_factor,
                        market.market_id,
                    )
                    self._runner_removals.append(_removal)
                    runner_removals.append(_removal)

        for _removal in runner_removals:
            self._process_runner_removal(market, *_removal)

        market.context["simulated"] = market_analytics
        # process simulated orders
        if market.blotter.active:
            self._process_simulated_orders(market, market_analytics)

    def remove_market(self, market) -> None:
        try:
            del self.markets[market.market_id]
        except KeyError:
            pass

    def _process_runner_removal(
        self,
        market,
        removal_selection_id: int,
        removal_handicap: int,
        removal_adjustment_factor: float,
    ) -> None:
        for order in market.blotter:
            if order.simulated:
                if order.lookup == (
                    market.market_id,
                    removal_selection_id,
                    removal_handicap,
                ):
                    # cancel and void order
                    order.simulated.size_matched = 0
                    order.simulated.average_price_matched = 0
                    order.simulated.matched = []
                    if order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
                        order.simulated.size_voided = order.order_type.size
                    else:
                        order.simulated.size_voided = order.order_type.liability
                    logger.warning(
                        "Order voided on non runner %s",
                        order.selection_id,
                        extra=order.info,
                    )
                else:
                    # TODO: "Where an SP lay bet in a win market has a maximum odds limit specified,..."
                    if (
                        order.order_type.ORDER_TYPE == OrderTypes.MARKET_ON_CLOSE
                    ) and order.side == "LAY":
                        if market.market_type == "WIN":
                            runner = [
                                x
                                for x in market.market_book.runners
                                if x.selection_id == order.selection_id
                                and x.handicap == order.handicap
                            ][0]
                            runner_adjustment_factor = runner.adjustment_factor
                            # See https://github.com/betcode-org/flumine/issues/454
                            multiplier = 1 - (
                                removal_adjustment_factor
                                / (100 - runner_adjustment_factor)
                            )
                            order.order_type.liability *= multiplier
                            if order.average_price_matched:
                                # We will get here if the NR is declared inplay
                                order.current_order.size_matched = round(
                                    order.order_type.liability
                                    / (order.average_price_matched - 1),
                                    2,
                                )
                            logger.warning(
                                "WIN MARKET_ON_CLOSE Order adjusted due to non runner %s",
                                order.selection_id,
                                extra=order.info,
                            )
                        elif market.market_type in {"PLACE", "OTHER_PLACE"}:
                            multiplier = (100 - removal_adjustment_factor) * 0.01
                            order.order_type.liability *= multiplier
                            if order.average_price_matched:
                                # We will get here if the NR is declared inplay
                                order.current_order.size_matched = round(
                                    order.order_type.liability
                                    / (order.average_price_matched - 1),
                                    2,
                                )
                            logger.warning(
                                "PLACE MARKET_ON_CLOSE Order adjusted due to non runner %s",
                                order.selection_id,
                                extra=order.info,
                            )
                    elif (
                        removal_adjustment_factor
                        and removal_adjustment_factor >= WIN_MINIMUM_ADJUSTMENT_FACTOR
                    ):
                        # todo place market
                        for match in order.simulated.matched:
                            match[1] = self._calculate_reduction_factor(
                                match[1], removal_adjustment_factor
                            )
                        _, order.simulated.average_price_matched = wap(
                            order.simulated.matched
                        )
                        logger.warning(
                            "Order adjusted due to non runner %s",
                            order.selection_id,
                            extra=order.info,
                        )

    @staticmethod
    def _calculate_reduction_factor(price: float, adjustment_factor: float) -> float:
        price_adjusted = round(price * (1 - (adjustment_factor / 100)), 2)
        return max(price_adjusted, 1.01)  # min: 1.01

    def _process_simulated_orders(self, market, market_analytics: dict) -> None:
        """
        #538 smart matching
          - isolation per order
            Potential double counting of passive liquidity, old logic no longer implemented
          - isolation per strategy (default)
            Prevent double counting of passive liquidity per strategy
          - isolation per instance
            Prevent double counting of passive liquidity on all orders regardless of strategy (interaction across strategies)
        """
        # isolation per strategy (default)
        if config.simulated_strategy_isolation:
            for strategy, orders in market.blotter._strategy_orders.items():
                live_orders = [
                    o for o in orders if o.status in LIVE_STATUS and o.simulated
                ]
                if live_orders:
                    _lookup = {
                        k: (v.runner, v.traded.copy())
                        for k, v in market_analytics.items()
                    }
                    live_orders_sorted = self._sort_orders(live_orders)
                    for order in live_orders_sorted:
                        runner_traded = _lookup[(order.selection_id, order.handicap)]
                        order.simulated(market.market_book, runner_traded)
        else:  # isolation per instance
            live_orders = list(market.blotter.live_orders)
            if live_orders:
                _lookup = {
                    k: (v.runner, v.traded.copy()) for k, v in market_analytics.items()
                }
                live_orders_sorted = self._sort_orders(live_orders)
                for order in live_orders_sorted:
                    if order.status in LIVE_STATUS and order.simulated:
                        runner_traded = _lookup[(order.selection_id, order.handicap)]
                        order.simulated(market.market_book, runner_traded)

    @staticmethod
    def _sort_orders(orders: list) -> list:
        # order by betId (default), side (Lay,Back) and then price
        lay_orders = sorted(
            [
                o
                for o in orders
                if o.side == "LAY"
                and o.order_type.ORDER_TYPE != OrderTypes.MARKET_ON_CLOSE
            ],
            key=lambda x: -x.order_type.price,
        )
        back_orders = sorted(
            [
                o
                for o in orders
                if o.side == "BACK"
                and o.order_type.ORDER_TYPE != OrderTypes.MARKET_ON_CLOSE
            ],
            key=lambda x: x.order_type.price,
        )
        moc = [
            o for o in orders if o.order_type.ORDER_TYPE == OrderTypes.MARKET_ON_CLOSE
        ]
        return lay_orders + back_orders + moc

    @staticmethod
    def _process_runner(market_analytics: dict, runner: RunnerBook) -> None:
        try:
            runner_analytics = market_analytics[(runner.selection_id, runner.handicap)]
        except KeyError:
            runner_analytics = market_analytics[
                (runner.selection_id, runner.handicap)
            ] = RunnerAnalytics(runner)
        runner_analytics(runner)


class RunnerAnalytics:
    def __init__(self, runner: RunnerBook):
        self.runner = runner
        self.traded = {}  # price: size traded since last update
        self._traded_volume = runner.ex.traded_volume
        self._p_v = {
            i["price"]: i["size"] for i in runner.ex.traded_volume
        }  # cached current volume

    def __call__(self, runner: RunnerBook):
        _tv = runner.ex.traded_volume
        if self._traded_volume == _tv:
            self.traded = {}
        else:
            self.traded = self._calculate_traded(_tv)
            self._traded_volume = _tv
        self.runner = runner

    def _calculate_traded(self, traded_volume: list) -> dict:
        p_v, traded = self._p_v, {}
        # create dictionary
        c_v = {i["price"]: i["size"] for i in traded_volume}
        # calculate difference
        for key, value in c_v.items():
            if key in p_v:
                new_value = float(value) - float(p_v[key])
                if new_value > 0:
                    traded[key] = round(new_value, 2)
            else:
                traded[key] = value
        # cache for next update
        self._p_v = c_v
        return traded


class SimulatedSportsDataMiddleware(Middleware):
    """
    Middleware to allow simulation of historic
    sports data.
    Creates generator of sports data that is cycled
    chronologically with the MarketBooks calling
    `strategy.process_sports_data`
    """

    def __init__(self, operation: str, directory: str):
        self.operation = operation  # "cricketSubscription" or "raceSubscription"
        self.directory = directory  # sports data directory (marketId used for lookup)
        self._gen = None
        self._next = None

    def __call__(self, market) -> None:
        pt = market.market_book.publish_time_epoch
        while True:
            if self._next is None:
                break
            for update in self._next:
                if update.market_id != market.market_id:
                    continue
                if pt > update.publish_time_epoch:
                    for strategy in market.flumine.strategies:
                        if (
                            market.market_book.streaming_unique_id
                            in strategy.stream_ids
                        ):
                            if call_strategy_error_handling(
                                strategy.check_sports_data, market, update
                            ):
                                call_strategy_error_handling(
                                    strategy.process_sports_data, market, update
                                )
                else:
                    return
            try:
                self._next = next(self._gen)
            except StopIteration:
                break

    def add_market(self, market) -> None:
        # create sports data generator
        file_path = os.path.join(self.directory, market.market_id)
        self._gen = self._create_generator(file_path, self.operation, 123)()
        try:
            self._next = next(self._gen)
        except StopIteration:
            logger.error(
                f"File {market.market_id} cannot be processed (data is not valid)"
            )
            self._next = None

    def remove_market(self, market) -> None:
        # clear gens
        self._gen = None
        self._next = None

    @staticmethod
    def _create_generator(file_path: str, operation: str, unique_id: int):
        listener = HistoricListener(max_latency=None, update_clk=False)
        stream = FlumineHistoricalGeneratorStream(
            file_path=file_path,
            listener=listener,
            operation=operation,
            unique_id=unique_id,
        )
        return stream.get_generator()



import logging

from ..clients.clients import VenueType
from ..order.ordertype import OrderTypes
from ..order.orderpackage import OrderPackageType, BaseOrder
from . import BaseControl
from .. import utils, config
from ..streams.orderstream import OrderStream

logger = logging.getLogger(__name__)


class OrderValidation(BaseControl):
    """
    Validates order price and size is valid for
    venue.
    """

    NAME = "ORDER_VALIDATION"

    def _validate(self, order: BaseOrder, package_type: OrderPackageType) -> None:
        if order.VENUE == VenueType.BETFAIR:
            self._validate_betfair_order(order)
        elif order.VENUE == VenueType.BETDAQ:
            self._validate_betdaq_order(order)

    def _validate_betfair_order(self, order):
        if order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
            self._validate_size(order)
            self._validate_betfair_price(order)
            self._validate_betfair_min_size(order, OrderTypes.LIMIT)
        elif order.order_type.ORDER_TYPE == OrderTypes.LIMIT_ON_CLOSE:
            self._validate_betfair_price(order)
            self._validate_betfair_liability(order)
            self._validate_betfair_min_size(order, OrderTypes.LIMIT_ON_CLOSE)
        elif order.order_type.ORDER_TYPE == OrderTypes.MARKET_ON_CLOSE:
            self._validate_betfair_liability(order)
            self._validate_betfair_min_size(order, OrderTypes.MARKET_ON_CLOSE)
        else:
            self._on_error(order, "Unknown orderType")

    def _validate_betdaq_order(self, order):
        if order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
            self._validate_size(order)
            self._validate_betdaq_price(order)
            self._validate_betdaq_min_size(order, OrderTypes.LIMIT)
        else:
            self._on_error(order, "Unknown orderType")

    def _validate_size(self, order):
        if order.VENUE == VenueType.BETFAIR:
            size = order.order_type.size or order.order_type.bet_target_size
        else:
            size = order.order_type.size
        if size is None:
            self._on_error(order, "Order size is None")
        elif size <= 0:
            self._on_error(order, "Order size is less than 0")
        elif size != round(size, 2):
            self._on_error(order, "Order size has more than 2dp")

    def _validate_betfair_price(self, order):
        if order.order_type.price is None:
            self._on_error(order, "Order price is None")
        if order.order_type.price_ladder_definition == "CLASSIC":
            if utils.as_dec(order.order_type.price) not in utils.PRICES:
                self._on_error(order, "Order price is not valid for CLASSIC ladder")
        elif order.order_type.price_ladder_definition == "FINEST":
            if utils.as_dec(order.order_type.price) not in utils.FINEST_PRICES:
                self._on_error(order, "Order price is not valid for FINEST ladder")
        elif order.order_type.price_ladder_definition == "LINE_RANGE":
            prices = utils.make_line_prices(
                order.order_type.line_range_info.min_unit_value,
                order.order_type.line_range_info.max_unit_value,
                order.order_type.line_range_info.interval,
            )
            if utils.as_dec(order.order_type.price) not in prices:
                self._on_error(order, "Order price is not valid for LINE_RANGE ladder")

    def _validate_betfair_liability(self, order):
        if order.order_type.liability is None:
            self._on_error(order, "Order liability is None")
        elif order.order_type.liability <= 0:
            self._on_error(order, "Order liability is less than 0")
        elif order.order_type.liability != round(order.order_type.liability, 2):
            self._on_error(order, "Order liability has more than 2dp")

    def _validate_betfair_min_size(self, order, order_type):
        client = order.client
        if client.min_bet_validation is False:
            return  # some accounts do not have min bet restrictions
        if order_type == OrderTypes.LIMIT:
            size = order.order_type.size or order.order_type.bet_target_size
            if (
                size < client.min_bet_size
                and (order.order_type.price * size) < client.min_bet_payout
            ):
                self._on_error(
                    order,
                    "Order size is less than min bet size ({0}) or payout ({1}) for currency".format(
                        client.min_bet_size, client.min_bet_payout
                    ),
                )
        else:  # todo is this correct?
            if (
                order.side == "BACK"
                and order.order_type.liability < client.min_bet_size
            ):
                self._on_error(
                    order,
                    "Liability is less than min bet size ({0}) for currency".format(
                        client.min_bet_size
                    ),
                )
            elif (
                order.side == "LAY"
                and order.order_type.liability < client.min_bsp_liability
            ):
                self._on_error(
                    order,
                    "Liability is less than min BSP payout ({0}) for currency".format(
                        client.min_bsp_liability
                    ),
                )

    def _validate_betdaq_price(self, order):
        if order.order_type.price is None:
            self._on_error(order, "Order price is None")
        # check ladder
        if utils.as_dec(order.order_type.price) not in utils.BETDAQ_PRICES:
            self._on_error(order, "Order price is not valid for BETDAQ ladder")

    def _validate_betdaq_min_size(self, order, order_type):
        client = order.client
        if client.min_bet_validation is False:
            return  # some accounts do not have min bet restrictions
        if order.order_type.size < 0.10:
            self._on_error(
                order,
                "Bet size is less than min bet size (0.10) for currency",
            )


class MarketValidation(BaseControl):
    """
    Validates market is available and open for orders
    """

    NAME = "MARKET_VALIDATION"

    def _validate(self, order: BaseOrder, package_type: OrderPackageType) -> None:
        if order.VENUE == VenueType.BETFAIR:
            self._validate_betfair_market_status(order)
        # todo BETDAQ

    def _validate_betfair_market_status(self, order):
        market = self.flumine.markets.markets.get(order.market_id)
        if market is None:
            self._on_error(order, "Market is not available")
        elif market.market_book is None:
            self._on_error(order, "MarketBook is not available")
        elif market.market_book.status != "OPEN":
            self._on_error(order, "Market is not open")


class ExecutionValidation(BaseControl):
    """
    Validates the OrderStream is connected when executing bets
    """

    NAME = "EXECUTION_VALIDATION"

    @property
    def order_stream_connected(self):
        for stream in self.flumine.streams:
            # todo handle multi clients / OrderStream
            if isinstance(stream, OrderStream):
                if stream.stream_running:
                    return True
        return False

    @staticmethod
    def failed_execution_attempts(responses):
        return len([response for response in responses if response.status != "SUCCESS"])

    def validate_order(self, order, package_type):
        if package_type == OrderPackageType.REPLACE:
            failed_attempts = self.failed_execution_attempts(
                order.responses.replace_responses
            )
        elif package_type == OrderPackageType.UPDATE:
            failed_attempts = self.failed_execution_attempts(
                order.responses.update_responses
            )
        elif package_type == OrderPackageType.CANCEL:
            failed_attempts = self.failed_execution_attempts(
                order.responses.cancel_responses
            )
        else:
            return

        if (
            not self.order_stream_connected
            and failed_attempts >= config.execution_retry_attempts
        ):
            self._on_error(
                order,
                "OrderStream is not connected, execution of orders is blocked until OrderStream is reconnected",
            )

    def _validate(self, order: BaseOrder, package_type: OrderPackageType) -> None:
        if self.flumine.clients.simulated:
            return
        if order.VENUE == VenueType.BETFAIR:
            self.validate_order(order, package_type)


class StrategyExposure(BaseControl):
    """
    Validates:
        - `strategy.validate_order` function
        - `strategy.max_order_exposure` is not violated if order is executed
        - `strategy.max_selection_exposure` is not violated if order is executed
        - `strategy.max_market_exposure` is not violated if order is executed

    Exposure calculation includes pending,
    executable and execution complete orders.
    """

    NAME = "STRATEGY_EXPOSURE"

    def _validate(self, order: BaseOrder, package_type: OrderPackageType) -> None:
        if package_type == OrderPackageType.PLACE:
            # strategy.validate_order
            runner_context = order.trade.strategy.get_runner_context(*order.lookup)
            if order.trade.strategy.validate_order(runner_context, order) is False:
                return self._on_error(order, order.violation_msg)

        if package_type in (
            OrderPackageType.PLACE,
            OrderPackageType.REPLACE,
        ) or (
            order.VENUE == VenueType.BETDAQ and package_type == OrderPackageType.UPDATE
        ):
            strategy = order.trade.strategy
            market = self.flumine.markets.markets[order.market_id]
            if (
                strategy.max_order_exposure is not None
                or strategy.max_selection_exposure is not None
            ):
                if order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
                    if order.order_type.price_ladder_definition == "LINE_RANGE":
                        # All bets are struck at 2.0
                        order_exposure = (
                            order.order_type.size or order.order_type.bet_target_size
                        )
                    else:
                        size = order.order_type.size or order.order_type.bet_target_size
                        price = order.order_type.price
                        if order.side == "BACK":
                            order_exposure = size
                        else:
                            order_exposure = (price - 1) * size
                elif order.order_type.ORDER_TYPE == OrderTypes.LIMIT_ON_CLOSE:
                    order_exposure = order.order_type.liability
                elif order.order_type.ORDER_TYPE == OrderTypes.MARKET_ON_CLOSE:
                    order_exposure = order.order_type.liability
                else:
                    return self._on_error(order, "Unknown order_type")

            # per order
            if (
                strategy.max_order_exposure is not None
                and order_exposure > strategy.max_order_exposure
            ):
                return self._on_error(
                    order,
                    "Order exposure ({0}) is greater than strategy.max_order_exposure ({1})".format(
                        order_exposure, strategy.max_order_exposure
                    ),
                )

            # per selection
            if strategy.max_selection_exposure is not None:
                if package_type == OrderPackageType.REPLACE:
                    exclusion = order
                else:
                    exclusion = None

                current_exposures = market.blotter.get_exposures(
                    strategy, lookup=order.lookup, exclusion=exclusion
                )
                """
                We use -min(...) in the below, as "worst_possible_profit_on_X" will be negative if the position is
                at risk of loss, while exposure values are always atleast zero.
                Exposure refers to the largest potential loss.
                """
                if order.side == "BACK":
                    current_selection_exposure = -current_exposures[
                        "worst_possible_profit_on_lose"
                    ]
                else:
                    current_selection_exposure = -current_exposures[
                        "worst_possible_profit_on_win"
                    ]
                potential_exposure = current_selection_exposure + order_exposure
                if potential_exposure > strategy.max_selection_exposure:
                    return self._on_error(
                        order,
                        "Potential selection exposure ({0:.2f}) is greater than strategy.max_selection_exposure ({1})".format(
                            potential_exposure,
                            strategy.max_selection_exposure,
                        ),
                    )

            # per market
            if strategy.max_market_exposure is not None:
                potential_exposure = -market.blotter.market_exposure(
                    strategy=strategy,
                    market_book=market.market_book,
                    exclusion=(
                        order if package_type == OrderPackageType.REPLACE else None
                    ),
                    new_order=order,
                )
                if potential_exposure > strategy.max_market_exposure:
                    return self._on_error(
                        order,
                        "Potential market exposure ({0:.2f}) is greater than strategy.max_market_exposure ({1})".format(
                            potential_exposure,
                            strategy.max_market_exposure,
                        ),
                    )



import logging
from typing import Optional, Union
from betfairlightweight import BetfairError, resources
from betfairlightweight.metadata import currency_parameters

from .baseclient import BaseClient
from .clients import VenueType

logger = logging.getLogger(__name__)

# default to GBP on error
MIN_BET_SIZE = currency_parameters["GBP"]["min_bet_size"]
MIN_BSP_LIABILITY = currency_parameters["GBP"]["min_bsp_liability"]
MIN_BET_PAYOUT = currency_parameters["GBP"]["min_bet_payout"]


class BetfairClient(BaseClient):
    """
    Betfair betting client.
    """

    VENUE = VenueType.BETFAIR

    def login(self) -> Optional[resources.LoginResource]:
        try:
            if self.interactive_login:
                return self.betting_client.login_interactive()
            else:
                return self.betting_client.login()
        except BetfairError as e:
            logger.error(
                "BetfairClient `login` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "login",
                    "response": e,
                },
            )

    def keep_alive(self) -> Optional[Union[resources.KeepAliveResource, bool]]:
        if self.betting_client.session_expired:
            try:
                return self.betting_client.keep_alive()
            except BetfairError as e:
                logger.error(
                    "BetfairClient `keep_alive` error",
                    exc_info=True,
                    extra={
                        "client": self.betting_client,
                        "trading_function": "keep_alive",
                        "response": e,
                    },
                )
        else:
            return True

    def logout(self) -> Optional[resources.LogoutResource]:
        try:
            return self.betting_client.logout()
        except BetfairError as e:
            logger.error(
                "BetfairClient `logout` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "logout",
                    "response": e,
                },
            )

    def update_account_details(self) -> None:
        # get details
        account_details = self._get_account_details()
        if account_details:
            self.account_details = account_details
        # get funds
        account_funds = self._get_account_funds()
        if account_funds:
            self.account_funds = account_funds

    def _get_account_details(self) -> Optional[resources.AccountDetails]:
        try:
            return self.betting_client.account.get_account_details()
        except BetfairError as e:
            logger.error(
                "BetfairClient `account.get_account_details` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "account.get_account_details",
                    "response": e,
                },
            )

    def _get_account_funds(self) -> Optional[resources.AccountFunds]:
        try:
            return self.betting_client.account.get_account_funds()
        except BetfairError as e:
            logger.error(
                "BetfairClient `account.get_account_funds` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "account.get_account_funds",
                    "response": e,
                },
            )

    @property
    def min_bet_size(self) -> Optional[float]:
        if self.account_details:
            try:
                return currency_parameters[self.account_details.currency_code][
                    "min_bet_size"
                ]
            except KeyError:
                logger.warning(
                    "min_bet_size KeyError: %s", self.account_details.currency_code
                )
                return MIN_BET_SIZE
            except Exception as e:
                logger.error("min_bet_size error: %s", e, exc_info=True)
                return MIN_BET_SIZE
        else:
            return MIN_BET_SIZE

    @property
    def min_bet_payout(self) -> Optional[float]:
        if self.account_details:
            try:
                return currency_parameters[self.account_details.currency_code][
                    "min_bet_payout"
                ]
            except KeyError:
                logger.warning(
                    "min_bet_payout KeyError: %s" % self.account_details.currency_code
                )
                return MIN_BET_PAYOUT
            except Exception as e:
                logger.error("min_bet_payout error: %s", e, exc_info=True)
                return MIN_BET_PAYOUT
        else:
            return MIN_BET_PAYOUT

    @property
    def min_bsp_liability(self) -> Optional[float]:
        if self.account_details:
            try:
                return currency_parameters[self.account_details.currency_code][
                    "min_bsp_liability"
                ]
            except KeyError:
                logger.warning(
                    "min_bsp_liability KeyError: %s", self.account_details.currency_code
                )
                return MIN_BSP_LIABILITY
            except Exception as e:
                logger.error("min_bsp_liability error: %s", e, exc_info=True)
                return MIN_BSP_LIABILITY
        else:
            return MIN_BSP_LIABILITY



import time
import logging
from betdaq import BetdaqError
from tenacity import retry

from .basestream import BaseStream
from ..clients import VenueType
from ..events.events import CurrentOrdersEvent

logger = logging.getLogger(__name__)

RETRY_WAIT = BaseStream.RETRY_WAIT
START_DELAY = 2
SNAP_DELTA = 2


class BetdaqOrderPolling(BaseStream):
    @retry(wait=RETRY_WAIT)
    def run(self) -> None:
        logger.info(
            f"Starting BetdaqOrderPolling '{self.stream_id}'",
            extra={
                "stream_id": self.stream_id,
                "client_username": self.client.username,
            },
        )
        # bootstrap call first to get most recent sequence number and all current orders
        order_count, sequence_number = 251, 0
        while order_count > 250:  # continue to call until less than 250 orders returned
            bootstrap = self.betting_client.betting.get_orders()
            current_orders = bootstrap["orders"]
            sequence_number = bootstrap["maximum_sequence_number"]
            self._process_current_orders(current_orders, sequence_number)
            order_count = len(current_orders)

        while self.is_alive():
            try:
                current_orders = self.betting_client.betting.get_orders_diff(
                    SequenceNumber=sequence_number
                )
            except (BetdaqError, Exception):
                logger.error(
                    "BetdaqOrderPolling %s run error", self.stream_id, exc_info=True
                )
                time.sleep(SNAP_DELTA)
                continue
            sequence_number = self._process_current_orders(
                current_orders, sequence_number
            )
            if self.flumine.markets.live_orders:
                time.sleep(self.streaming_timeout)
            else:
                time.sleep(SNAP_DELTA)

        logger.info(
            f"Stopped BetdaqOrderPolling '{self.stream_id}'",
            extra={
                "stream_id": self.stream_id,
                "client_username": self.client.username,
            },
        )

    def _process_current_orders(
        self, current_orders: list, sequence_number: int
    ) -> int:
        if current_orders or self.flumine.markets.live_orders:
            self.flumine.handler_queue.put(
                CurrentOrdersEvent(current_orders, venue=VenueType.BETDAQ)
            )
            # update SequenceNumber
            for order in current_orders:
                sequence_number = max(order["sequence_number"], sequence_number)
        return sequence_number

    @property
    def stream_running(self) -> bool:
        return True



import logging
from typing import Optional
from betconnect.exceptions import BetConnectException
from betconnect import resources

from .baseclient import BaseClient
from .clients import VenueType

logger = logging.getLogger(__name__)


class BetConnectClient(BaseClient):
    """
    BetConnect betting client.
    """

    VENUE = VenueType.BETCONNECT

    def login(self) -> Optional[resources.Login]:
        try:
            return self.betting_client.account.login()
        except BetConnectException as e:
            logger.error(
                "BetConnectClient `account.login` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "account.login",
                    "response": e,
                },
            )

    def keep_alive(self) -> Optional[resources.Login]:
        try:
            return self.betting_client.account.refresh_session_token()
        except BetConnectException as e:
            logger.error(
                "BetConnectClient `account.refresh_session_token` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "account.refresh_session_token",
                    "response": e,
                },
            )

    def logout(self) -> None:
        try:
            return self.betting_client.account.logout()
        except BetConnectException as e:
            logger.error(
                "BetConnectClient `account.logout` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "account.logout",
                    "response": e,
                },
            )

    def update_account_details(self) -> None:
        # get details
        account_details = self._get_account_details()
        if account_details:
            self.account_details = account_details
        # get funds
        account_funds = self._get_account_funds()
        if account_funds:
            self.account_funds = account_funds

    def _get_account_details(self) -> Optional[resources.AccountPreferences]:
        try:
            return self.betting_client.account.get_user_preferences()
        except BetConnectException as e:
            logger.error(
                "BetConnectClient `account.get_user_preferences` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "account.get_user_preferences",
                    "response": e,
                },
            )

    def _get_account_funds(self) -> Optional[resources.Balance]:
        try:
            return self.betting_client.account.get_balance()
        except BetConnectException as e:
            logger.error(
                "BetConnectClient `account.get_balance` error",
                exc_info=True,
                extra={
                    "client": self.betting_client,
                    "trading_function": "account.get_balance",
                    "response": e,
                },
            )



import threading
import queue
import logging
import betfairlightweight
from betfairlightweight import StreamListener
from tenacity import wait_exponential

logger = logging.getLogger(__name__)


class BaseStream(threading.Thread):
    LISTENER = StreamListener
    MAX_LATENCY = 0.5
    RETRY_WAIT = wait_exponential(multiplier=1, min=2, max=60)

    def __init__(
        self,
        flumine,
        stream_id: int = None,
        streaming_timeout: float = None,  # snaps listener if no update
        conflate_ms: int = None,
        market_filter: dict = None,
        market_data_filter: dict = None,
        sports_data_filter: str = None,
        custom: bool = False,
        client=None,
        output_queue: bool = True,
        event_processing: bool = False,
        event_group: str = None,
        event_id: str = None,
        operation: str = "marketSubscription",
        **listener_kwargs,
    ):
        threading.Thread.__init__(self, daemon=True, name=self.__class__.__name__)
        self.flumine = flumine
        self.stream_id = stream_id
        self.market_filter = market_filter
        self.market_data_filter = market_data_filter
        self.sports_data_filter = sports_data_filter
        self.streaming_timeout = streaming_timeout
        self.conflate_ms = conflate_ms
        self._client = client
        self.custom = custom
        self._stream = None
        self._output_queue = queue.Queue() if output_queue else None
        self.event_processing = event_processing
        self.event_group = event_group
        self.event_id = event_id
        self.operation = operation
        self.listener_kwargs = listener_kwargs
        self._listener = self.LISTENER(
            output_queue=self._output_queue,
            max_latency=self.MAX_LATENCY,
            **listener_kwargs,
        )
        self._output_thread = threading.Thread(
            name="{0}_output_thread".format(self.name),
            target=self.handle_output,
            daemon=True,
        )

    def run(self) -> None:
        raise NotImplementedError

    def handle_output(self) -> None:
        raise NotImplementedError

    def stop(self) -> None:
        if self._stream:
            self._stream.stop()

    @property
    def betting_client(self) -> betfairlightweight.APIClient:
        return self.client.betting_client

    @property
    def client(self):
        if self._client:
            return self._client
        else:
            return self.flumine.clients.get_default()

    @property
    def stream_running(self) -> bool:
        if self._stream:
            return self._stream.running
        else:
            return False



import datetime
from typing import Optional


class Responses:
    """Order responses"""

    def __init__(self):
        self.date_time_created = datetime.datetime.now(datetime.timezone.utc)
        self.current_order = None  # resources.CurrentOrder
        self.place_response = None  # resources.PlaceOrderInstructionReports
        self.cancel_responses = []
        self.replace_responses = []
        self.update_responses = []
        self._date_time_placed = None

    def placed(self, response=None, dt: bool = True) -> None:
        if response:
            self.place_response = response
        if dt:
            self._date_time_placed = datetime.datetime.now(datetime.timezone.utc)

    def cancelled(self, response) -> None:
        self.cancel_responses.append(response)

    def replaced(self, response) -> None:
        self.replace_responses.append(response)

    def updated(self, response) -> None:
        self.update_responses.append(response)

    @property
    def date_time_placed(self) -> Optional[datetime.datetime]:
        if self._date_time_placed:
            return self._date_time_placed
        elif self.current_order:
            try:
                return self.current_order.placed_date
            except AttributeError:
                return
